#Case Studies
from class_game import *
import ast
from visual_automata.fa.dfa import VisualDFA
from visual_automata.fa.nfa import VisualNFA


import numpy as np

### Initialization function
def Init(Mat_transition,Istate,Fstate):
  #State init
  state_i=[]
  N=len(Mat_transition)
  M=len(Mat_transition[0])
  for nn in range(N):
    state_i.append("q"+str(nn))
  state_i=set(state_i)

  #Name_Transitions
  list_trans_i=list(dict.fromkeys(Mat_transition.flatten()))
  trans_i=[]
  for ti in list_trans_i:
    if ti:
      trans_i.append(str(ti))
  trans_i=set(trans_i)
  Istate="q"+str(Istate-1)
  FS=[]
  for kk in range(len(Fstate)):
    FS.append("q"+str(Fstate[kk]-1))
  if len(FS)==1:
    Fstate={FS}
  else:
    Fstate=set(FS)
  dict_trans={}
  for id_state in range(N):
    dict_state={}
    for id_transition in range(M):
      value=Mat_transition[id_state][id_transition]
      if value:
        out="q"+str(id_transition)
        dict_state[str(value)]={out}
    dict_trans["q"+str(id_state)]=dict_state
  #return state_i,trans_i,dict_trans,Istate,Fstate
  nfa=VisualNFA(
      states=state_i,
      input_symbols=trans_i,
      transitions=dict_trans,
      initial_state=Istate,
      final_states=Fstate)

  table=nfa.table
  graph=nfa.show_diagram()
  return graph,table,nfa




#### Dining Cryptographe

def dining_crypt(ab="Head",ac="Tail",bc="Head",who_paid=[0,0,0]):
  dining=game(3,["A","B","C"]) 
  ab=ab
  ac=ac
  bc=bc
  list_value=[ac,ab,bc]
  #mat=[[0,ab,ac],[ab,0,bc],[ac,bc,0]]
  mat=[[0,ab,ac],[0,0,0],[0,bc,0]]
  dining.make_transition(mat)

  for id in range(len(list_value)):
    if list_value[id]=="Head":
      list_value[id]=1
    else:
      list_value[id]=0
  list_XOR=[]
  for kk in range(3):
    if not(who_paid[kk]):
      list_XOR.append(list_value[kk]^list_value[(kk+1)%3])
    else:
      list_XOR.append(not(list_value[kk]^list_value[(kk+1)%3]))
    
  list_XOR_label=[]
  for XOR in list_XOR:
    if XOR:
      list_XOR_label.append("True")
    else:
      list_XOR_label.append("False")
  rez=list_XOR[0]^list_XOR[1]^list_XOR[2]
  f=dining.display_diagram(list_XOR_label)
  if rez==0:
    if sum(who_paid)==0:
      f.attr(label='Result:NSA paid. Good Prediction',frontcolor='red')
    else:
      f.attr(label='Result:NSA paid. Bad Prediction')
  else:
    if sum(who_paid)==0:
     f.attr(label='Result: Someone paid. Bad Prediction')
    else:
      f.attr(label='Result: Someone paid. Good Prediction')
  
  return f


### ICGS train

def iCGS_train(ActT1="r",ActT2="r",ActC="6",boolean_Red=True,cs1='a',cs2='a',cs3='o',cs4='o',cs5='e',cs6='e'):
  #icgs: M=< Ag={t1,t2,c}, AP={p,b,d}, S={sI,s1,s2,s3,s4,s5,s6,s7?}, so=SI, Actt1={r,l,i}, Act2={r,l,s,i},Actc={1,2,3,4,5,6,a,e,o,i}> #We swich Agent and State
  test=game(load_file=True,AW=True)
  if ActC=='6' and ActT1=='r' and ActT2=='r':
    if cs6=='e':
      SP=3
    elif cs6=='o':
      SP=7
    else:
      SP=6
  elif ActC=='2' and ActT1=='l' and ActT2=='l':
    if cs2=='o':
      SP=7
    else:
      SP=2
  elif ActC=='3' and ActT1=='r' and ActT2=='l':
    if cs3=='e':
      SP=2
    elif cs3=='o':
      SP=7
    else:
      SP=3
  elif ActC=='4' and ActT1=='r' and ActT2=='s':
    if cs4=='a':
      SP=0
    elif cs4=='o':
      SP=7
    else:
      SP=4
  elif ActC=='5' and ActT1=='l' and ActT2=='r':
    if cs5=='a':
      SP=0
    elif cs5=='o':
      SP=7
    else:
      SP=5
  elif ActC=='1' and ActT1=='l' and ActT2=='s':
    if cs1=='a':
      SP=0
    elif cs1=='o':
      SP=7
    else:
      SP=1
  else:
    SP=0
  if boolean_Red:
    graph=test.display_diagram(Special_node=SP)
  else:
    graph=test.display_diagram()
  return graph


### Example Graph
def example_strategy(h1,h2):
  game_strategy=game(load_file=True,path1='example_strategy.txt',AW=True)
  if h1[0]=='A' and h2[0]!='0':
    if h2[1]=='D':
      if (h1[2]=='A' and h2[2]=='D') or (h1[2]=='B' and h2[2]=='C'):
        SP=3
      else:
        SP=2
    elif h2[1]=='C':
      SP=3
    else:
      SP=1
  elif h1[0]=='B' and h2[0]!='0':
    if (h1[2]=='A' and h2[2]=='D') or (h1[2]=='B' and h2[2]=='C'):
      SP=3
    else:
      SP=2
  else:
    SP=0
  graph=game_strategy.display_diagram(Special_node=SP)
  return graph


def display_graph_MS(Mat_transition,List_name_agent,List_name_stat):
  test=game(len(List_name_stat),List_name_stat,AW=True)
  test.make_transition(Mat_transition)
  test=test.display_diagram()
  return test





import sys
from vitamin_model_checker.models.costCGS.costCGS import *
from binarytree import Node
from vitamin_model_checker.logics.RBATL import *


# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cCGS.get_matrix_proposition()

    index = cCGS.get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
    return states



# set of states (ex. s1, s2) as input and returns a set of indices to identify them
def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = cCGS.get_index_by_state_name(elem)
        states.add(int(position))
    return states


# converts a string into a set
def string_to_set(string):
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)


#  function that builds a formula tree, used by the model checker
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        states = set()
        states_proposition = get_states_prop_holds(str(tpl))
        if states_proposition is None:
            return None
        else:
            for element in states_proposition:
                states.add(cCGS.get_state_name_by_index(element))
            root = Node(str(states))
    return root

# # Compute the cost of the joint action (RBATL)
# def Cost(actions, state, coalition):
#     total = sys.maxsize
#     for action in actions:
#         max_action = 0
#         for action1 in actions:
#             for i in coalition:
#                 if action[int(i)-1] != action1[int(i)-1]:
#                     break
#             else:
#                 max_action = max(get_cost_for_action(action1, 's'+str(state)), max_action)
#         total = min(max_action, total)
#     return total

# Compute the cost of the joint action (RBATL)
def goodActionsCost(actions, state, bound):
    good_actions = set()
    for action in actions:
        if cCGS.get_cost_for_action(action, 's'+str(state)) <= bound:
            good_actions.add(action)
    return good_actions
        
# It returns the states from which the coalition has a strategy to enforce the next state to lie in state_set.
# function used by the model checker.
def pre(coalition, state_set, bound):
    agents = cCGS.get_agents_from_coalition(coalition)
    graph = cCGS.get_graph()
    state_set = convert_state_set(state_set)  # returns a set of indexes
    pre_states = set()
    dict_state_action = dict()  # dictionary state-action
    for i, source in enumerate(graph):  # take states that have at least one transition to one of the states in the set
        for j in state_set:
            if graph[i][j] != 0:
                coordinates = str(i) + "," + str(j)
                actions_list = cCGS.build_list(graph[i][j])
                goodActions = goodActionsCost(actions_list, i, bound)
                if goodActions:
                    dict_state_action.update({coordinates: goodActions})

    # iterate over these states and check that my move is not present in any other state
    for key, value in dict_state_action.items():
        other_actions_in_row = dict()  # dictionary containing other moves of the row (excluding those saved in dict_state_action).
        all_actions_in_row = set()  # all elements in the row
        i = int(key.split(',')[0])
        j = int(key.split(',')[1])
        for index, element in enumerate(graph[i]):
            if element != 0:
                coordinates = str(i) + "," + str(index)
                all_actions_in_row.update(cCGS.build_list(graph[i][index]))
                if index != j:
                    other_actions_in_row.update({coordinates: cCGS.build_list(graph[i][index])})

        # check if there is a loop -> if yes, it is a pre
        for y in value:
            if '*' in y:
                pre_states.add(str(i))
                break

        for action in value:
            move = cCGS.get_coalition_action(set([action]), agents)
            # checks whether a move is present in the others
            check_passed = True
            for k, v in other_actions_in_row.items():
                other_act = cCGS.get_coalition_action(v, agents)
                if move.intersection(other_act): # they have something in common
                    # check if it belongs to the set
                    column = int(k.split(',')[1])
                    if column not in state_set:
                        check_passed = False
            if check_passed == True:
                pre_states.add(i)

    # take the elements of the set and check if the moves towards the state contain all the possible opponent moves
    result = set()
    for i in pre_states:
        i = int(i)
        opponent_moves_in_state = set()
        opponent_moves_in_row = set()
        for j, column in enumerate(graph[int(i)]):
            if graph[i][j] != 0:
                opponent_moves_in_row.update(cCGS.get_opponent_moves(cCGS.build_list(graph[i][j]), agents))

                if j in state_set:
                    # takes the opponent's moves for actions towards the states in state_set
                    opponent_moves_in_state.update(cCGS.get_opponent_moves(cCGS.build_list(graph[i][j]), agents))

        if opponent_moves_in_row == opponent_moves_in_state:
            # convert i into corresponding state
            result.add(cCGS.get_state_name_by_index(i))
    return result

def extract_coalition_and_cost(string):
    tmp = string[1:].split(">")
    coalition = tmp[0]
    cost = [int(i) for i in tmp[1][1:].split(',')]
    return (coalition, cost)

def inc_bound(bound_p, bound):
    for i in range(0, len(bound_p)):
        if bound[i] == 0:
            bound_p[i] = 0
        else:
            bound_p[i] = (bound_p[i]+1) % (bound[i]+1)
        if bound_p[i] != 0:
            break
def diff_bound(bound1, bound2):
    bound = [0]*len(bound1)
    for i in range(0, len(bound1)):
        bound[i] = bound1[i]-bound2[i]
    return bound

solve_tree_cache = dict()

# function that solves the formula tree. The result is the model checking result.
# It solves every node depending on the operator.
def solve_tree(node):
    key = str(node.value) + (str(node.left) if node.left is not None else '') + (str(node.right) if node.right is not None else '')
    if key in solve_tree_cache:
        node.value = solve_tree_cache[key]
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)

    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            all_states = set(cCGS.get_states())
            ris = all_states - states
            node.value = str(ris)

        elif verify('COALITION_BOUND', node.value) and verify('GLOBALLY', node.value):  # e.g. <1>Gφ
            coalition_and_bound = extract_coalition_and_cost(node.value)
            coalition = coalition_and_bound[0] # extract the coalition of agents
            bound = coalition_and_bound[1] # extract the bound b assigned to the strategies
            states = string_to_set(node.left.value)
            if not any(bound):
                p = set(cCGS.get_states())
                t = states
                while p - t:  # p not in t
                    p = t
                    t = pre(coalition, p, bound) & states
                node.value = str(p)
            else:
                p = set()
                t = set()
                value_backup = node.value
                bound_p = [0]*len(bound)
                while True:
                    node.value = value_backup[:value_backup[1:].index('<')+1] + '<' + ','.join([str(b) for b in bound_p]) + '>G'
                    solve_tree(node)
                    t = pre(coalition, string_to_set(node.value), diff_bound(bound, bound_p)) & states
                    while p - t:  # p not in t
                        p.update(t)
                        t = pre(coalition, p, [0]*len(bound)) & states
                    inc_bound(bound_p, bound)
                    if bound_p == bound: break
                node.value = str(p)            

        elif verify('COALITION_BOUND', node.value) and verify('NEXT', node.value):  # e.g. <1>Xφ
            # coalition = node.value[1:-2]
            coalition_and_bound = extract_coalition_and_cost(node.value)
            coalition = coalition_and_bound[0] # extract the coalition of agents
            bound = coalition_and_bound[1] # extract the bound b assigned to the strategies
            states = string_to_set(node.left.value)
            ris = pre(coalition, states, bound)
            node.value = str(ris)

        elif verify('COALITION_BOUND', node.value) and verify('EVENTUALLY', node.value):  # e.g. <1>Fφ
            # trueUϕ.
            # coalition = node.value[1:-2]
            coalition_and_bound = extract_coalition_and_cost(node.value)
            coalition = coalition_and_bound[0] # extract the coalition of agents
            bound = coalition_and_bound[1] # extract the bound b assigned to the strategies
            states1 = set(cCGS.get_states())
            states2 = string_to_set(node.left.value)
            if not any(bound):
                p = set()
                t = states2
                while t - p:  # t not in p
                    p.update(t)
                    t = pre(coalition, p, bound) & states1
                node.value = str(p)
            else:
                p = set()
                t = set()
                value_backup = node.value
                bound_p = [0]*len(bound)
                while True:
                    node.value = value_backup[:value_backup[1:].index('<')+1] + '<' + ','.join([str(b) for b in bound_p]) + '>F'
                    solve_tree(node)
                    t = pre(coalition, string_to_set(node.value), diff_bound(bound, bound_p)) & states1
                    while t - p:  # t not in p
                        p.update(t)
                        t = pre(coalition, p, [0]*len(bound)) & states1
                    inc_bound(bound_p, bound)
                    if bound_p == bound: break
                node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)

        elif verify('COALITION_BOUND', node.value) and verify('UNTIL', node.value):  # e.g. <1>φUθ
            # coalition = node.value[1:-2]
            coalition_and_bound = extract_coalition_and_cost(node.value)
            coalition = coalition_and_bound[0] # extract the coalition of agents
            bound = coalition_and_bound[1] # extract the bound b assigned to the strategies
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            if not any(bound):
                p = set()
                t = states2
                while t - p:  # t not in p
                    p.update(t)
                    t = pre(coalition, p, bound) & states1
                node.value = str(p)
            else:
                p = set()
                t = set()
                value_backup = node.value
                bound_p = [0]*len(bound)
                while True:
                    node.value = value_backup[:value_backup[1:].index('<')+1] + '<' + ','.join([str(b) for b in bound_p]) + '>U'
                    solve_tree(node)
                    t = pre(coalition, string_to_set(node.value), diff_bound(bound, bound_p)) & states1
                    while t - p:  # t not in p
                        p.update(t)
                        t = pre(coalition, p, [0]*len(bound)) & states1
                    inc_bound(bound_p, bound)
                    if bound_p == bound: break
                node.value = str(p)
            
        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cCGS.get_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)

    if key not in solve_tree_cache:
        solve_tree_cache[key] = node.value

# returns whether the result of model checking is true or false in the initial state
def verify_initial_state(initial_state, string):
    if initial_state in string:
        return True
    return False


# does the parsing of the model, the formula, builds a tree and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cCGS

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cCGS = costCGS()
    cCGS.read_file(filename)

    # formula parsing
    res_parsing = do_parsing(formula, cCGS.get_number_of_agents())
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result

    # model checking
    solve_tree(root)

    # solution
    initial_state = cCGS.get_initial_state()
    bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value), 'initial_state': 'Initial state '+ str(initial_state) + ": " + str(bool_res)}
    return result





from vitamin_model_checker.model_checker_interface.explicit.ATLF.pre_ATLF import *
from binarytree import Node
from vitamin_model_checker.logics.ATL import *
from vitamin_model_checker.models.CGS import *


# returns a list of tuples (state, value) where the value depends on the input proposition
def get_tuple_list_prop(prop):
    i = cgs.get_atom_index(prop)
    if i is None:
        return None
    states = cgs.get_states()
    list = []
    matrix = cgs.get_matrix_proposition()

    for index, source in enumerate(matrix):
        tuple = (states[index], float(source[i]))
        list.append(tuple)
    return list


# function that builds a formula tree, used by the model checker
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        couples = get_tuple_list_prop(tpl)
        if couples is None:
            return None
        root = Node(str(couples))

    return root


# converts a string in a list of tuple (state, value)
def string_to_tuple_list(string):
    list = string.strip("[]").split(", ")
    new_list = []
    for i in range(0, len(list), 2):
        state = list[i].strip("('")
        value = float(list[i + 1].strip(")"))
        new_tuple = (state, value)
        new_list.append(new_tuple)
    return new_list


## returns the minimum between matching values
# e.g. min(states1[2], states2[2])
def intersection_values(states1, states2):
    list = []
    for i in range(0, len(states1)):
        value = min(states1[i][1], states2[i][1])
        tuple = (states1[i][0], value)
        list.append(tuple)
    return list


# assign a value to all tuples (state, value) in the list
def set_value_tuple_list(value):
    list = []
    states = cgs.get_states()
    for i in range(0, len(states)):
        tuple = (states[i], value)
        list.append(tuple)
    return list


# returns if p is in t
def first_included_in_second(p, t):
    # p and t are list of tuples
    for i in range(0, len(p)):
        value_p = p[i][1]
        value_t = t[i][1]
        if value_p > value_t:
            return False
    return True


# returns a new list of tuples with the value of t if value_t is bigger than value_p,
# with the value of p otherwise
def update_values(p, t):
    list = []
    for i in range(0, len(p)):
        value_p = p[i][1]
        value_t = t[i][1]
        if value_t > value_p:
            tuple = (p[i][0], value_t)
        else:
            tuple = (p[i][0], value_p)
        list.append(tuple)
    return list


# returns the difference between matching values
# e.g. states1[2] - states2[2]
def difference_values(states1, states2):
    list = []
    for i in range(0, len(states1)):
        value = states1[i][1] - states2[i][1]
        tuple = (states1[i][0], value)
        list.append(tuple)
    return list


# returns the max between matching values
# e.g. max(states1[2], states2[2])
def union_values(states1, states2):
    list = []
    for i in range(0, len(states1)):
        value = max(states1[i][1], states2[i][1])
        tuple = (states1[i][0], value)
        list.append(tuple)
    return list


# function that solves the formula tree. The result is the model checking result.
# It solves every node depending on the operator.
def solve_tree(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)

    if node.right is None:  # UNARY OPERATORS: not, globally, next, eventually
        # ¬φ := 1−φ
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_tuple_list(node.left.value)
            ris = difference_values(set_value_tuple_list(1), states)
            node.value = str(ris)

        if verify('COALITION', node.value) and verify('GLOBALLY', node.value):  # e.g. <1>Gφ
            coalition = node.value[1:-2]
            states = string_to_tuple_list(node.left.value)
            p = set_value_tuple_list(1)  # true
            t = states
            while not first_included_in_second(p, t):  # p not in t
                p = t
                t = intersection_values(eval(pre(cgs,coalition, p)), states)
            node.value = str(p)

        elif verify('COALITION', node.value) and verify('NEXT', node.value):  # e.g. <1>Xφ
            coalition = node.value[1:-2]
            ris = pre(cgs,coalition, eval(node.left.value))
            node.value = str(ris)

        elif verify('COALITION', node.value) and verify('EVENTUALLY', node.value):  # e.g. <1>Fφ
            coalition = node.value[1:-2]
            states = string_to_tuple_list(node.left.value)
            p = set_value_tuple_list(0)
            t = states
            while not first_included_in_second(t, p):  # t not in p
                p = update_values(p, t)
                t = eval(pre(cgs,coalition, p))
            node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value):  # e.g. φ || θ
            states1 = string_to_tuple_list(node.left.value)
            states2 = string_to_tuple_list(node.right.value)
            res = []
            for i in range(0, len(states1)):
                state = states1[i][0]
                value = max(states1[i][1], states2[i][1])
                tuple = (state, value)
                res.append(tuple)
            node.value = str(res)

        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_tuple_list(node.left.value)
            states2 = string_to_tuple_list(node.right.value)
            res = []
            for i in range(0, len(states1)):
                state = states1[i][0]
                value = min(states1[i][1], states2[i][1])
                tuple = (state, value)
                res.append(tuple)
            node.value = str(res)

        elif verify('COALITION', node.value) and verify('UNTIL', node.value): # e.g. <1>φUθ
            coalition = node.value[1:-2]
            states1 = string_to_tuple_list(node.left.value)
            states2 = string_to_tuple_list(node.right.value)
            p = set_value_tuple_list(0)
            t = states2
            while not first_included_in_second(t, p):  # t not in p
                p = update_values(p, t)
                t = intersection_values(eval(pre(cgs,coalition, p)), states1)
            node.value = str(p)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_tuple_list(node.left.value)
            states2 = string_to_tuple_list(node.right.value)
            not_states1 = difference_values(set_value_tuple_list(1), states1)
            ris = union_values(not_states1, states2)
            node.value = str(ris)


# returns the value of the model checking in the initial state
def get_value_initial_state(initial_state, string):
    list_tuple = eval(string)
    for element in list_tuple:
        if element[0] == initial_state:
            return element[1]


# does the parsing of the model, the formula, builds a tree, and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cgs

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cgs = CGS()
    cgs.read_file(filename)

    # formula parsing
    res_parsing = do_parsing(formula, cgs.get_number_of_agents())
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result

    # model checking
    solve_tree(root)

    #solution
    initial_state = cgs.get_initial_state()
    value_initial_state = get_value_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value),
              'initial_state': 'Initial state ' + str(initial_state) + ": " + str(value_initial_state)}
    return result



from vitamin_model_checker.models.costCGS.costCGS import *
from binarytree import Node
from vitamin_model_checker.logics.OATL import *
from timeit import default_timer as timer


# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cCGS.get_matrix_proposition()

    index = cCGS.get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
    return states



# set of states (ex. s1, s2) as input and returns a set of indices to identify them
def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = cCGS.get_index_by_state_name(elem)
        states.add(int(position))
    return states


# converts a string into a set
def string_to_set(string):
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)
    
    
def complement(state_set):
    result = set()
    graph = cCGS.get_graph()
    if (len(graph) - len(state_set) <= 0): 
        return result
    for i in range(0, len(graph)):
        if (i not in state_set):
            result.add(i)
    return result


#  function that builds a formula tree, used by the model checker
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        states = set()
        if (verify('FALSE', str(tpl))):
            return Node(str(states))
        elif (verify('TRUE', str(tpl))):
            return Node(str(set(cCGS.get_states())))
        states_proposition = get_states_prop_holds(str(tpl))
        if states_proposition is None:
            return None
        else:
            for element in states_proposition:
                states.add(cCGS.get_state_name_by_index(element))
            root = Node(str(states))
    return root



def pre(state_set):
    state_set = convert_state_set(state_set)  # returns a set of indexes
    graph = cCGS.get_graph()
    result = set()
    for state in state_set:
        for i in range(0, len(graph)):
                if graph[i][state] != 0:
                    result.add(cCGS.get_state_name_by_index(i))
    return result
    

def check_if_action_is_extension(action, extension_action):
    for index, letter in enumerate(action):
        if letter == '-':
            continue
        elif letter != extension_action[index]:
            return False
    return True
    
def get_action_extensions(action, actions_at_state):
    result = set()
    for extension_action in actions_at_state:
        if extension_action == action:
            continue
        if check_if_action_is_extension(action, extension_action):
            result.add(extension_action)
    return result        
    
def next(action, state):
    graph = cCGS.get_graph()
    result = set()
    for index, elem in enumerate(graph[state]):
        if graph[state][index] != 0:
            for elem_action in cCGS.build_list(elem):
                if action == elem_action or check_if_action_is_extension(action, elem_action):
                    result.add(index)
                    break
    return result         
        

def D(s, agents, state_set):
    agents = cCGS.get_agents_from_coalition(agents)
    tmp_set = set()
    tmp_set.add(s)
    state_index = convert_state_set(tmp_set).pop()
    state_set_index = convert_state_set(state_set)
    state_set_complement = complement(state_set_index)
    graph = cCGS.get_graph()
    result = set()
    for index, elem in enumerate(graph[state_index]):
        if graph[state_index][index] != 0:
            for action in cCGS.build_list(graph[state_index][index]):
                base_action = cCGS.get_base_action(action, agents)
                if (next(base_action, state_index) <= state_set_complement):
                    result.add(action)
    return result
    
def Cost(action_set, state):
    total = 0
    for action in action_set:
        total += cCGS.get_cost_for_action(action, state)
    return total
    
def cross(n, agents, states):
    result = set()
    for state in pre(states):
        actions = D(state, agents, states)
        if (Cost(actions, state) <= n):
            result.add(state)
    return result
    
def extract_coalition_and_cost(string):
    tmp = string[1:].split(">")
    coalition = tmp[0]
    cost = tmp[1][1:]
    return (coalition, cost)
    
    
# function that solves the formula tree. The result is the model checking result.
# It solves every node depending on the operator.
def solve_tree(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)

    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            all_states = set(cCGS.get_states())
            ris = all_states - states
            node.value = str(ris)

        elif verify('COALITION_DEMONIC', node.value) and verify('GLOBALLY', node.value):  # e.g. <Jn>Gφ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states1 = set()
            states2 = string_to_set(node.left.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | cross(n, coalition, p))
            node.value = str(p)

        elif verify('COALITION_DEMONIC', node.value) and verify('NEXT', node.value):  # e.g. <i,j><n>Xφ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states = string_to_set(node.left.value)
            ris = cross(n, states)
            node.value = str(ris)

        elif verify('COALITION_DEMONIC', node.value) and verify('EVENTUALLY', node.value):  # e.g. <i,j><n>Fφ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states1 = set(cCGS.get_states())
            states2 = string_to_set(node.left.value)
            p = set()
            t = states2
            while t != p:
                p = t
                t = states2 | (states1 & cross(n, coalition, p))
            node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)

        elif verify('COALITION_DEMONIC', node.value) and verify('UNTIL', node.value):  # e.g. <i,j><n>φUθ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set()
            t = states2
            while t != p: 
                p = t
                t = states2 | (states1 & cross(n, coalition, p))
            node.value = str(p)
        elif verify('COALITION_DEMONIC', node.value) and verify('RELEASE', node.value): #e.g. <i,j><n>φRθ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | cross(n, coalition, p))
            node.value = str(p)
        elif verify('COALITION_DEMONIC', node.value) and verify('WEAK', node.value): #e.g. <i,j><n>φWθ
            coalition_and_cost = extract_coalition_and_cost(node.value)
            coalition = coalition_and_cost[0]
            n = int(coalition_and_cost[1])
            states1 = string_to_set(node.right.value)
            states2 = string_to_set(node.left.value) | states1
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | cross(n, coalition, p))
            node.value = str(p)
        
        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cCGS.get_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)

# returns whether the result of model checking is true or false in the initial state
def verify_initial_state(initial_state, string):
    if initial_state in string:
        return True
    return False


# does the parsing of the model, the formula, builds a tree and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cCGS

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cCGS = costCGS()
    cCGS.read_file(filename)

    # formula parsing
    res_parsing = do_parsing(formula, cCGS.get_number_of_agents())
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result

    # model checking
    solve_tree(root)

    # solution
    initial_state = cCGS.get_initial_state()
    bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value), 'initial_state': 'Initial state '+ str(initial_state) + ": " + str(bool_res)}
    return result
    
    
def model_checking_test(formula, filename):
    global cCGS

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cCGS = costCGS()
    cCGS.read_file(filename)

    # formula parsing
    res_parsing = do_parsing(formula, cCGS.get_number_of_agents())
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result

    # model checking
    time1 = timer()
    solve_tree(root)
    model_checking_time = timer() - time1
    # solution
    initial_state = cCGS.get_initial_state()
    bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value), 'initial_state': 'Initial state '+ str(initial_state) + ": " + str(bool_res), 'time': model_checking_time}
    return result





import graphviz

### Game Class
class game:
  def __init__(self,Ag=1,name_list=[],load_file=False,path1="", AW=False):
    self.AW=AW
    if load_file:
      if path1=="":
        Mat_tansition,Mat_unknow,State,IState=self.load()
      else:
        Mat_tansition,Mat_unknow,State,IState=self.load(path1)
      self.name_list=State
      self.Ag=len(Mat_tansition)
      self.list_agent=[]
      self.build()
      self.make_transition(Mat_tansition)
      self.Unknow(Mat_unknow)
    else:
      self.name_list=name_list
      self.Ag=Ag
      self.list_agent=[]
      self.build()

  
  class agent:
    def __init__(self,name):
      self.name=name
      self.list_transition=[]
      self.unkown_transition=[]
    def make_agent_transition(self,list_transi):
      self.list_transition=list_transi
    def make_agent_UT(self,list_transi):
      self.unkown_transition=list_transi

  def build(self):
    self.list_agent=[]
    for ag in range(self.Ag):
      if self.name_list==[]:
        self.list_agent.append(self.agent(str(ag)))
      else:
        self.list_agent.append(self.agent(self.name_list[ag]))
  def make_transition(self,Mat_transition):
    for id_agent in range(len(Mat_transition)):
      (self.list_agent[id_agent]).make_agent_transition(Mat_transition[id_agent])
  def Unknow(self,Mat):
    for id_agent in range(len(Mat)):
      (self.list_agent[id_agent]).make_agent_UT(Mat[id_agent])

  def diagram_init(self,Special_node=-1):
    f = graphviz.Digraph('finite_state_machine',{'arrowhead' : 'none'})
    f.attr('node', shape='circle')
    f.attr('edge',{'arrowhead':'none'})
    if self.AW:
      f.attr('edge',{'arrowhead':'normal'})
    if Special_node!=-1:
      f.node(self.list_agent[Special_node].name,style='filled', fillcolor='red')
    for id_agent in range(self.Ag):
      for id_transition in range(len(self.list_agent[id_agent].list_transition)):
        if type(self.list_agent[id_agent].list_transition[id_transition])==str and self.list_agent[id_agent].list_transition[id_transition] !='No Action':
          f.edge(self.list_agent[id_agent].name,self.list_agent[id_transition].name,label=str(self.list_agent[id_agent].list_transition[id_transition]))
    return f

  def diagram_unk(self,col,f):
    f.attr('edge',{'arrowhead' :'none', 'style' :'dotted'})
    for id_agent in range(self.Ag):
      for id_transition in range(len(self.list_agent[id_agent].unkown_transition)):
        if type(self.list_agent[id_agent].unkown_transition[id_transition])==str and self.list_agent[id_agent].unkown_transition[id_transition] !='0':
          f.edge(self.list_agent[id_agent].name,self.list_agent[id_transition].name,color=col,label=str(self.list_agent[id_agent].unkown_transition[id_transition]))
    f.attr('edge', {'arrowhead' :'Normal', 'style':'solid'})
    return f

  def label_node(self,f,list_label):
    for ag in range(self.Ag):
      f.node(str(self.list_agent[ag].name),xlabel=str(list_label[ag]))
    return f

  def display_diagram(self,list_label=[],col='blue',mat_add=False,Special_node=-1):
    f=self.diagram_init(Special_node)
    if not(mat_add):
      f=self.diagram_unk(col,f)
    if list_label==[]:
      return f
    else:
      return self.label_node(f,list_label)


  def store(self, path): 

    with open(path) as file:
      file.write('''
Transition
0 HH HM HL 0 0 0 0 0 0 0 0 0
0 0 0 0 MM ML 0 0 0 0 0 0 0
0 0 0 0 0 0 MH ML 0 0 0 0 0
0 0 0 0 0 0 0 0 MH MM 0 0 0
0 0 0 0 0 0 0 0 0 0 LL 0 0
0 0 0 0 0 0 0 0 0 0 LM 0 0
0 0 0 0 0 0 0 0 0 0 LL 0 0
0 0 0 0 0 0 0 0 0 0 0 LH 0
0 0 0 0 0 0 0 0 0 0 0 0 LM
0 0 0 0 0 0 0 0 0 0 LH 0 0
0 0 0 0 0 0 0 0 0 0 * 0 0
0 0 0 0 0 0 0 0 0 0 0 * 0
0 0 0 0 0 0 0 0 0 0 0 0 *
{transitions}
Name_State
s0 s1 s2 s3 s4 s5 s6 s7 s8 s9 tie winK winT
Initial_State
s0
Atomic_propositions
kingwin tianjiwin tie kingnotwin tianjinotwin
Labelling
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 0 1 1
0 0 1 1 1
1 0 0 0 1
0 1 0 1 0
Number_of_agents
2
''')

  def load(self,path='example_train_controller.txt'):
    file=open(path, 'r', encoding = 'utf-8-sig')
    lines=file.readlines()
    file.close()
    Mat_tansition=[]
    Mat_unknow=[]
    IState=[]
    State=[]
    cmpt=0
    for line in lines:
      if line=="Transition\n" or line=="Unkown_Transition_by\n" or line=="Name_State\n" or line=="Initial_State\n" or line=="":
        cmpt+=1
      else:
        if cmpt==1:
          line = line.split(' ')
          line = [i.strip() for i in line]
          Mat_tansition.append(line)
        elif cmpt==2:
          line = line.split(' ')
          line = [i.strip() for i in line]
          Mat_unknow.append(line)
        elif cmpt==3:
          line = line.split(' ')
          State = [i.strip() for i in line]
        elif cmpt==4:
          line = line.split(' ')
          IState=[i.strip() for i in line]
          cmpt+=1
    return Mat_tansition,Mat_unknow,State,IState



from vitamin_model_checker.model_checker_interface.explicit.CapATL.CapATL import *

# returns possible actions given a set of capacities i.e. [('c','c','c1'),('c', 'c', 'c2')]
# it can corresponds to a knowledge delta
def get_actions_from_capacity_set(capacity_set) :
    result = []
    a = []
    for elem in capacity_set :
        for cap in elem :
            a.append(get_actions_from_capacity(cap))
        a = trouver_combinaisons(a)
        for x in a : 
            if x not in result :
                result.append(x)
        a = []
    return result


def group_by_state(succ_w):
    states = []
    interm = []
    ajout = []
    intermediaire_state = 0
    for success in succ_w :
        intermediaire_state = getattr(success, 'state')
        if intermediaire_state in states :
            for elem in interm :
                if elem[0] == intermediaire_state :
                    elem.append(success)
        if intermediaire_state not in states :
            states.append(intermediaire_state)
            ajout = [intermediaire_state, success]
            interm.append(ajout)
# interm is a list of succ group by their state 
# i.e. [[q1, succ1, succ2],[q3, succ1, succ2]]
    return interm


def action_elem12(elem1, elem2) :
    know_1 = elem1.knowledge
    know_2 = elem2.knowledge
    if tuple(elem1.state) != elem2.state :
        return False
    
    for j1, j2 in enumerate(know_1) :
        for q in j2 :
            if q not in know_2[j1] :
                return False 
    return elem2.action


def action_in_W2(elem, W) :
    result = []
    for a in W :
        if action_elem12(elem, a) != False : 
            result.append(action_elem12(elem, a))
    return result


# returns the intersection of actions i.e. to check if for a state q, unique couple state action
def intersection_same_q(group_by_elmt, W) :
    act = []
    result = []

    for elem in group_by_elmt[1:] :
        if action_in_W2(elem, W) != [] :
            act.append(action_in_W2(elem, W))
    for i in act :
        for j in i : 
            if j not in result :
                result.append(j)
    return result



def unique_state_action_couple(succw,W):
    w_succ = succw
    group = group_by_state(w_succ)
    for elem in group :
        if (len(intersection_same_q(elem,W))) != 1 :
            return False 
    return True


# returns if w=[q,delta1,...,deltak] is in W i.e. there existe an action beta' such that 
# [q,delta1,...,deltak, beta'] is in W
def elem_in_W(w, W) :
    result = []
    interm = []
    k = 0
    for value in W :
        interm = [ x for x in getattr(value, 'knowledge') if x not in getattr(w, 'knowledge')]
        if len(interm)==0 and getattr(value, 'state') == getattr(w, 'state'):
            k += 1
            result.append(value)
    if k == 0 :
        return None 
    return result


def succ_in_W(succw,W, dict_W) :
    
    k = 0 
    succ_w = succw

    l = len(succ_w)
    if l ==0 : 
        return False
    if succ_w == None :
        print("pb succ")
        return False
    for elem in succ_w :
        key_interm = (tuple(elem.state), (elem.knowledge))

        if key_interm not in dict_W : 
            return False
    return True 


def pre(W,coal_Y):
    P = []
    omega_Y = Omega_Y(coal_Y)
    dict_W = {}
    for j in W : 
        key = (j.state, j.knowledge)
        dict_W[key] = j
    for j in omega_Y :
        j1 = succ(j)
        if succ_in_W(j1,W, dict_W) and unique_state_action_couple(j1,W) :
            if j not in P :
                P.append(j)
    return P






import sys
sys.path.append('/media/angelo/WorkData/git/VITAMIN')
from vitamin_model_checker.model_checker_interface.explicit import RBATL, RABATL
from vitamin_model_checker.utils.generators import RBCGS_generator, RABCGS_generator
import time


def run_experiments(num_states_min, num_states_max, num_resources_min, num_resources_max, bound, repetitions):
    with open('results.csv', 'w') as file:
        file.write('# States; # Resources; RBATL time [sec]; RABATL time [sec]\n')    
    for state in range(num_states_min, num_states_max+1, 10):
        for resource in range(num_resources_min, num_resources_max+1):
            phi = '(<1,2><{res}>F p)'.format(res=','.join([str(bound) for _ in range(0, resource)]))
            avg_rbtime = 0
            avg_rabtime = 0
            with open('results.csv', 'a') as file:
                file.write(str(state) + ';' + str(resource) + ';')
            for _ in range(0, repetitions):
                RBCGS_generator.generate_random_model_file(state, resource, 'tmp.txt')
                start = time.time()
                RBATL.model_checking(phi, 'tmp.txt')
                end = time.time()
                avg_rbtime += end-start
                RABCGS_generator.generate_random_model_file(state, resource, 'tmp.txt')
                start = time.time()
                RABATL.model_checking(phi, 'tmp.txt')
                end = time.time()
                avg_rabtime += end-start
            avg_rbtime = avg_rbtime / repetitions
            avg_rbtime = avg_rabtime / repetitions
            with open('results.csv', 'a') as file:
                file.write(str(avg_rbtime) + '; ' + str(avg_rabtime) + '\n')

if __name__ == "__main__":
    num_states_min = max(int(sys.argv[1]), 1)
    num_states_max = max(int(sys.argv[2]), 1)
    num_resources_min = max(int(sys.argv[3]), 1)
    num_resources_max = max(int(sys.argv[4]), 1)
    bound = max(int(sys.argv[5]), 1)
    repetitions = max(int(sys.argv[6]), 1)
    run_experiments(num_states_min, num_states_max, num_resources_min, num_resources_max, bound, repetitions)





from xml_data.convertion import *

#an attack must be identified by it's id, an is composed of :
#An action
#A set of preconditions that the attacker must possess/know to launch the attack
#A postcondition that is the gain obtained by the attacker once the attack as succeded
class Attack:
    def __init__(self, id, action, preconditions, postcondition):
        self.id = id
        self.action = action
        self.preconditions = preconditions
        self.postcondition = postcondition
        

#take in input an xml file and output the 'mulval' hypergraph associated : the matrix of transition, a list of the attacks and a list of the conditions nodes
def initialisation(file):
    #We get the treated data from the xml
    (transition, node) = build_data(file)
    condition = []
    
    #The transition matrix
    nb_trans = len(transition)
    transition_matrix = [[0 for i in range(nb_trans)] for i in range(nb_trans)]
    for trans in transition:
        #We reduces the indexes by 1, as in the mulval xml it start at 0
        transition_matrix[trans.src - 1][trans.dst - 1] = 1
        
    #the attacks list
    attaques = []
    for noeud in node:
        #AND noes are nodes representing the attacks
        if noeud.type == 'AND':
            #we get the id by decreasing the index of 1. This way we have a correspondance with it's position in the transition matrix
            id = noeud.id
            #we get the post and pre condition with the transition matrix
            for i in range(nb_trans):
                if transition_matrix[id][i] == 1:
                    postcondition = [node[i].fact]
            preconditions = []
            for i in range(nb_trans):
                if transition_matrix[i][id] == 1:
                    preconditions.append(node[i].fact)
            action = node[id].fact
            
            attaques.append(Attack(id, action, preconditions, postcondition))
        else : 
            #other nodes are condition and caracterised only by the knowledge they represent
            condition.append(noeud)
            
    return (transition_matrix, attaques, condition)
            
    
        
    



from anytree import Node, RenderTree
from vitamin_model_checker.models.CGS import *

# returns a tree, given the action and the tree set.
def get_coalition_tree(action_coalition, trees):
    for t in trees:
        if t.name == action_coalition:
            return t
    return None


def print_tree(tree):
    for pre, fill, node in RenderTree(tree):
        print(f"{pre}{node.name}")


def build_node(strategy, state, parent):
    node = [state, str(strategy)]
    return node

# returns the atom value in a given state
def get_atom_value(atom_values, state):
    for elem in atom_values:
        if elem[0] == state:
            return elem[1]


# returns a set of trees, one for each action. They represent the next possible move.
def create_next_move_trees(cgs, agents, state):
    destination = 0
    graph = cgs.get_graph()

    # transitions from a state
    index_state = cgs.get_index_by_state_name(state)
    trees = []
    for action in graph[index_state]:
        if action != 0:
            action = cgs.build_list(action)
            for move in action:
                coalition_move = cgs.get_coalition_action(set([move]), agents)

                # create a tree for each action
                t = get_coalition_tree(coalition_move, trees)
                opponent_move = cgs.get_opponent_moves(action, agents)

                if t is None:
                    tree = Node(coalition_move)
                    node = build_node(opponent_move, destination, coalition_move)
                    Node(node, parent=tree)
                    trees.append(tree)

                else:
                    node = build_node(opponent_move, destination, t.name)
                    Node(node, parent=t)

            # ---- end for move ----
        destination = destination + 1
    # ---- end for action ----
    return trees


# evaluates max(min(t1, t2, t3...))
def evaluate_max_strategy(cgs, agents, state, atom_values):

    # builds the strategy tree and then calculates the minimum value
    min_values_tree = []
    trees = create_next_move_trees(cgs, agents, state)

    for el in trees:
        children = el.children
        values_tree = []
        for node in children:
            state = cgs.get_state_name_by_index(node.name[0])
            values_tree.append(get_atom_value(atom_values, state))

        minimum = min(values_tree)
        min_values_tree.append(minimum)

    return max(min_values_tree)

# It returns the states from which the coalition has a strategy to enforce the next state to lie in state_set.
# function used by the model checker.
def pre(cgs, coalition, atom_values):
    # pre for each state
    agents = cgs.get_agents_from_coalition(coalition)
    states = cgs.get_states()
    result = []
    for state in states:
        # coalition strategy from state
        max_value = evaluate_max_strategy(cgs, agents, state, atom_values)
        # put value in the state tuple
        tuple = (state, max_value)
        result.append(tuple)

    # ------ end for state ---------
    result = str(result)
    return result



from vitamin_model_checker.models.CGS import *
from binarytree import Node
from vitamin_model_checker.logics.CTL import verifyCTL, do_parsingCTL

# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cgs.get_matrix_proposition()

    index = cgs.get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
            #states.add("s" + str(state))
    return states


# set of states (ex. s1, s2) as input and returns a set of indices to identify them
def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = cgs.get_index_by_state_name(elem)
        states.add(int(position))
    return states


# converts a string into a set
def string_to_set(string):
    #print(f"string:{string}")
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)


#  function that builds a formula tree, used by the model checker
#This function builds a formula tree, creating nodes for operators and atomic propositions
# Eg: Input: !AXa, Tree Root: NOT operator, Left Child: AXa
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        states = set()
        states_proposition = get_states_prop_holds(str(tpl))
        if states_proposition is None:
            return None
        else:
            for element in states_proposition:
                states.add(cgs.get_state_name_by_index(element))
            root = Node(str(states))
    return root


# It returns the states predecessors of those held in input
def pre_image_exist(transitions, list_holds_p):
    #print(f"input transitions: {transitions}")
    #print(f"pre-image: here're the input states where prop holds {list_holds_p}")
    pre_list = set()
    for state in list(list_holds_p):
        predecessors = {s for s, t in transitions if t == state}
        #print(f"pre-image predecessors: {predecessors}")
        pre_list.update(predecessors)
    return pre_list

def pre_image_all(transitions, list_holds_p):
    pre_list = set()
    for state in list(list_holds_p):
        predecessors = {s for s, t in transitions if t == state}
        for predecessor in predecessors:
            successor_states = {t for s, t in transitions if s == predecessor}
            if successor_states.issubset(list_holds_p):
                pre_list.add(predecessor)
    return pre_list


#This function solves formula tree returning a result for model checking. It analyzes tree recursively, solving nodes depending on the associated operator.
# Eg: Same build tree formula: !AXa -> this function starts to solve AXa node as first, and then applies NOT at the result.
#To solve AXa node, it uses pre-image function to define the states set where a proposition is satisfied. The result will be assigned to AXa node.
#At the end, the algorithm applies NOT operator at the result for AXa. To do so, it computes the complementary states set for those that satisfy AXa.
# The final result is stored in NOT operator root.
def solve_tree(node):

    if node.left is not None:
        print(f"node left: {node.left}")
        print(type(node.left))
        solve_tree(node.left)


    if node.right is not None:
        print(f"node right: {node.right}")
        solve_tree(node.right)

    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verifyCTL('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            ris = set(cgs.get_states()) - states
            node.value = str(ris)

        elif verifyCTL('EXIST', node.value) and verifyCTL('GLOBALLY', node.value):  # e.g. EGφ
            states = string_to_set(node.left.value)
            #print(f"states labelled as node lef(states where prop holds): {states}")
            p = set(cgs.get_states())
            t = states
            while p - t:  # p not in t
                p = t
                t = pre_image_exist(cgs.get_edges(), p) & states
            node.value = str(p)

        elif verifyCTL('FORALL', node.value) and verifyCTL('GLOBALLY', node.value):  # not(EF(not p)) e.g. AGφ
            states = string_to_set(node.left.value)
            compl_states = set(cgs.get_states()) - states
            #print(f"states labelled as node lef(states where prop holds): {compl_states}")
            p = set()
            t = compl_states
            while t - p:  # t not in p
                p.update(t)
                t = pre_image_all(cgs.get_edges(), p)
            out = set(cgs.get_states()) - p
            node.value = str(out)

        elif verifyCTL('FORALL', node.value) and verifyCTL('NEXT', node.value):# e.g. EXφ
            states = string_to_set(node.left.value)
            negated_states = set(cgs.get_states()) - states
            #print(f"states labelled as node lef(states where prop holds): {negated_states}")
            ris = pre_image_all(cgs.get_edges(), negated_states)
            complement = set(cgs.get_states()) - ris
            print(f"compl: {complement}")
            node.value = str(complement)

        elif (verifyCTL('EXIST', node.value) and verifyCTL('NEXT', node.value)):
            states = string_to_set(node.left.value)
            #print(f"states labelled as node lef(states where prop holds): {states}")
            ris = pre_image_exist(cgs.get_edges(), states)
            node.value = str(ris)


        elif verifyCTL('EXIST', node.value) and verifyCTL('EVENTUALLY', node.value):  # trueUϕ.
            states = string_to_set(node.left.value)
            #print(f"states labelled as node lef(states where prop holds): {states}")
            p = set()
            t = states
            while t - p:  # t not in p
                p.update(t)
                t = pre_image_exist(cgs.get_edges(), p)
            node.value = str(p)

        elif verifyCTL('FORALL', node.value) and verifyCTL('EVENTUALLY', node.value):  # not (EG(not p))
            states = string_to_set(node.left.value)
            compl_states = set(cgs.get_states()) - states
            #print(f"states labelled as node lef(states where prop holds): {compl_states}")
            p = set(cgs.get_states())
            t = compl_states
            while p - t:  # p not in t
                p = t
                t = pre_image_all(cgs.get_edges(),p) & compl_states
            ris = set(cgs.get_states()) - p
            node.value = str(ris)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verifyCTL('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)


        elif verifyCTL('EXIST', node.value) and verifyCTL('UNTIL', node.value):  # e.g. AφUθ
            states1 = string_to_set(node.left.value)
            #print(f"states labelled as node left(states where prop holds): {states1}")
            states2 = string_to_set(node.right.value)
            #print(f"states labelled as node right(states where prop holds): {states2}")
            p = set()
            t = states2
            while t - p:  # t not in p
                p.update(t)
                t = pre_image_exist(cgs.get_edges(), p) & states1
            node.value = str(p)

        elif verifyCTL('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verifyCTL('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cgs.get_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)

# returns whether the result of model checking is true or false in the initial state
def verify_initial_state(initial_state, string):
    if initial_state in string:
        return True
    return False


# does the parsing of the model, the formula, builds a tree and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cgs

    if not formula.strip():
        result = {'res': 'Error: formula not entered', 'initial_state': ''}
        return result

    # model parsing
    cgs = CGS()
    cgs.read_file(filename)

    # formula parsing
    print(f"formula to parse: {formula}")
    res_parsing = do_parsingCTL(formula)
    if res_parsing is None:
        result = {'res': "Syntax Error", 'initial_state': ''}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist", 'initial_state': ''}
        return result
    print(f"root is: {root}")
    # model checking
    solve_tree(root)

    # solution
    initial_state = cgs.get_initial_state()
    bool_res = verify_initial_state(initial_state, root.value)
    result = {'res': 'Result: ' + str(root.value), 'initial_state': 'Initial state '+ str(initial_state) + ": " + str(bool_res)}
    return result

# def process_modelCheckingCTL(filename, formula):
#     cgs = CGS()
#     cgs.read_file(filename)
#     result = model_checking(formula, filename)
#     print(result)
#     return result









# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'AGENT AND COALITION IS KCAP LPAREN NEXT NOT PROP RELEASE RPAREN UNTILexpression : expression AND expressionexpression : NOT expressionexpression : COALITION expression UNTIL expression \n                    | COALITION expression RELEASE expressionexpression :  COALITION NEXT expressionexpression : LPAREN expression RPARENexpression : PROPexpression : KCAP AGENT expression2expression2 : LPAREN expression2 RPARENexpression2 : expression2 AND expression2expression2 : NOT expression2 expression2 : AGENT IS PROP'
    
_lr_action_items = {'NOT':([0,2,3,4,7,10,12,14,15,20,21,25,],[2,2,2,2,2,2,21,2,2,21,21,21,]),'COALITION':([0,2,3,4,7,10,14,15,],[3,3,3,3,3,3,3,3,]),'LPAREN':([0,2,3,4,7,10,12,14,15,20,21,25,],[4,4,4,4,4,4,20,4,4,20,20,20,]),'PROP':([0,2,3,4,7,10,14,15,24,],[5,5,5,5,5,5,5,5,28,]),'KCAP':([0,2,3,4,7,10,14,15,],[6,6,6,6,6,6,6,6,]),'$end':([1,5,8,13,16,17,19,22,23,27,28,29,30,],[0,-7,-2,-1,-5,-6,-8,-3,-4,-11,-12,-10,-9,]),'AND':([1,5,8,9,11,13,16,17,19,22,23,26,27,28,29,30,],[7,-7,7,7,7,7,7,-6,25,7,7,25,25,-12,25,-9,]),'NEXT':([3,],[10,]),'UNTIL':([5,8,9,13,16,17,19,22,23,27,28,29,30,],[-7,-2,14,-1,-5,-6,-8,-3,-4,-11,-12,-10,-9,]),'RELEASE':([5,8,9,13,16,17,19,22,23,27,28,29,30,],[-7,-2,15,-1,-5,-6,-8,-3,-4,-11,-12,-10,-9,]),'RPAREN':([5,8,11,13,16,17,19,22,23,26,27,28,29,30,],[-7,-2,17,-1,-5,-6,-8,-3,-4,30,-11,-12,-10,-9,]),'AGENT':([6,12,20,21,25,],[12,18,18,18,18,]),'IS':([18,],[24,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,3,4,7,10,14,15,],[1,8,9,11,13,16,22,23,]),'expression2':([12,20,21,25,],[19,26,27,29,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression AND expression','expression',3,'p_expression_binary','parser_ATLe.py',56),
  ('expression -> NOT expression','expression',2,'p_expression_not','parser_ATLe.py',60),
  ('expression -> COALITION expression UNTIL expression','expression',4,'p_expression_ternary','parser_ATLe.py',68),
  ('expression -> COALITION expression RELEASE expression','expression',4,'p_expression_ternary','parser_ATLe.py',69),
  ('expression -> COALITION NEXT expression','expression',3,'p_expression_unary','parser_ATLe.py',78),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','parser_ATLe.py',90),
  ('expression -> PROP','expression',1,'p_expression_prop','parser_ATLe.py',94),
  ('expression -> KCAP AGENT expression2','expression',3,'p_expression_kcap','parser_ATLe.py',99),
  ('expression2 -> LPAREN expression2 RPAREN','expression2',3,'p_expression_capformula_group','parser_ATLe.py',104),
  ('expression2 -> expression2 AND expression2','expression2',3,'p_expression_capformula_and','parser_ATLe.py',109),
  ('expression2 -> NOT expression2','expression2',2,'p_expression_capformula_not','parser_ATLe.py',114),
  ('expression2 -> AGENT IS PROP','expression2',3,'p_expression_capformula_is','parser_ATLe.py',119),
]




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'AND EVENTUALLY EXIST FORALL GLOBALLY IMPLIES LPAREN NEXT NOT OR PROP RPAREN UNTILexpression : expression AND expression\n                  | expression OR expression\n                  | expression IMPLIES expressionexpression : FORALL expression UNTIL expression\n                  | EXIST expression UNTIL expressionexpression : FORALL GLOBALLY expression\n                  | FORALL NEXT expression\n                  | FORALL EVENTUALLY expression\n                  | EXIST GLOBALLY expression\n                  | EXIST NEXT expression\n                  | EXIST EVENTUALLY expressionexpression : NOT expressionexpression : LPAREN expression RPARENexpression : PROP'
    
_lr_action_items = {'FORALL':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,]),'EXIST':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,]),'NOT':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,]),'LPAREN':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,5,]),'PROP':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,6,]),'$end':([1,6,18,20,21,22,24,25,26,28,29,30,31,32,33,],[0,-14,-12,-1,-2,-3,-6,-7,-8,-9,-10,-11,-13,-4,-5,]),'AND':([1,6,10,14,18,19,20,21,22,24,25,26,28,29,30,31,32,33,],[7,-14,7,7,7,7,7,7,7,7,7,7,7,7,7,-13,7,7,]),'OR':([1,6,10,14,18,19,20,21,22,24,25,26,28,29,30,31,32,33,],[8,-14,8,8,8,8,8,8,8,8,8,8,8,8,8,-13,8,8,]),'IMPLIES':([1,6,10,14,18,19,20,21,22,24,25,26,28,29,30,31,32,33,],[9,-14,9,9,9,9,9,9,9,9,9,9,9,9,9,-13,9,9,]),'GLOBALLY':([2,3,],[11,15,]),'NEXT':([2,3,],[12,16,]),'EVENTUALLY':([2,3,],[13,17,]),'UNTIL':([6,10,14,18,20,21,22,24,25,26,28,29,30,31,32,33,],[-14,23,27,-12,-1,-2,-3,-6,-7,-8,-9,-10,-11,-13,-4,-5,]),'RPAREN':([6,18,19,20,21,22,24,25,26,28,29,30,31,32,33,],[-14,-12,31,-1,-2,-3,-6,-7,-8,-9,-10,-11,-13,-4,-5,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expression':([0,2,3,4,5,7,8,9,11,12,13,15,16,17,23,27,],[1,10,14,18,19,20,21,22,24,25,26,28,29,30,32,33,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expression","S'",1,None,None,None),
  ('expression -> expression AND expression','expression',3,'p_expression_binary','CTLparser.py',42),
  ('expression -> expression OR expression','expression',3,'p_expression_binary','CTLparser.py',43),
  ('expression -> expression IMPLIES expression','expression',3,'p_expression_binary','CTLparser.py',44),
  ('expression -> FORALL expression UNTIL expression','expression',4,'p_expression_ternary','CTLparser.py',48),
  ('expression -> EXIST expression UNTIL expression','expression',4,'p_expression_ternary','CTLparser.py',49),
  ('expression -> FORALL GLOBALLY expression','expression',3,'p_expression_unary','CTLparser.py',53),
  ('expression -> FORALL NEXT expression','expression',3,'p_expression_unary','CTLparser.py',54),
  ('expression -> FORALL EVENTUALLY expression','expression',3,'p_expression_unary','CTLparser.py',55),
  ('expression -> EXIST GLOBALLY expression','expression',3,'p_expression_unary','CTLparser.py',56),
  ('expression -> EXIST NEXT expression','expression',3,'p_expression_unary','CTLparser.py',57),
  ('expression -> EXIST EVENTUALLY expression','expression',3,'p_expression_unary','CTLparser.py',58),
  ('expression -> NOT expression','expression',2,'p_expression_not','CTLparser.py',62),
  ('expression -> LPAREN expression RPAREN','expression',3,'p_expression_group','CTLparser.py',66),
  ('expression -> PROP','expression',1,'p_expression_prop','CTLparser.py',70),
]



import string
import time

import streamlit as st

# from logics.ATL import ATL
# from logics.ATLF import ATLF
# from logics.CapATL import CapATL
# from logics.OL import OL
# from logics.OATL import OATL
# from logics.NatATL import *
from back_end_CS import *
import os
from vitamin_model_checker.models import *
from pathlib import *
# from vitamin_model_checker.xml_data import *
# from vitamin_model_checker.xml_data.writting import writting


def display_case(nlp_steps):


  ### 01

  if nlp_steps == "01 - Initialization":
    st.markdown("---")
    st.write(f"        ")

    st.header("01 - Initialization")
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### i - Quick Introduction: ')

    st.write("""Game theory is the study of mathematical models of strategic interactions among rational agents.
    It has applications in all fields of social science, as well as in logic, systems science and computer science.
    Originally, it addressed two-person zero-sum games, in which each participant's gains or losses are exactly balanced by those of other participants.
    In the 21st century, game theory applies to a wide range of behavioral relations; it is now an umbrella term for the science of logical decision making in humans, animals, as well as computers.""")

    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### ii - Diagram: ')

    Mat_transition=np.array([[3,1,2,0,0],
    [0,3,1,2,0],
    [0,0,3,1,2],
    [0,0,0,3,1],
    [0,0,0,0,3]])

    Istate=1

    Fstate=[4,5]
    t1,t2,t3 = Init(Mat_transition,Istate,Fstate)
    st.graphviz_chart(t1)

    st.write(f"        ")




    snippet = f"""

    def Init(Mat_transition,Istate,Fstate):
    #State init
    state_i=[]
    N=len(Mat_transition)
    M=len(Mat_transition[0])
    for nn in range(N):
      state_i.append("q"+str(nn))
    state_i=set(state_i)

    #Name_Transitions
    list_trans_i=list(dict.fromkeys(Mat_transition.flatten()))
    trans_i=[]
    for ti in list_trans_i:
      if ti:
      trans_i.append(str(ti))
    trans_i=set(trans_i)
    Istate="q"+str(Istate-1)
    FS=[]
    for kk in range(len(Fstate)):
      FS.append("q"+str(Fstate[kk]-1))
    if len(FS)==1:
      Fstate=dic(FS)
    else:
      Fstate=set(FS)
    dict_trans=dic()
    for id_state in range(N):
      dict_state=dic()
      for id_transition in range(M):
      value=Mat_transition[id_state][id_transition]
      if value:
      out="q"+str(id_transition)
      dict_state[str(value)]=dic(out)
      dict_trans["q"+str(id_state)]=dict_state
    #return state_i,trans_i,dict_trans,Istate,Fstate
    nfa=VisualNFA(
      states=state_i,
      input_symbols=trans_i,
      transitions=dict_trans,
      initial_state=Istate,
      final_states=Fstate)

    table=nfa.table
    graph=nfa.show_diagram()
    return graph,table,nfa

    """
    code_header_placeholder = st.empty()
    snippet_placeholder = st.empty()

    st.write(f"        ")
    st.write(f"        ")
    code_header_placeholder.markdown(f"#### iii - Code 01 - Initialization")
    snippet_placeholder.code(snippet)
    st.write("     ")
    st.write("     ")
    st.markdown("---")




  ### 04
  if nlp_steps == "04 - Dining Cryptographers":
    st.markdown("---")
    st.write(f"        ")

    st.header("04 - Dining Cryptographers")
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### i - Quick Introduction: ')
    st.write(f"        ")
    st.write(f"        ")

    st.write(""" In cryptography, the dining cryptographers problem studies how to perform a secure multi-party computation of the boolean-XOR function.
  David Chaum first proposed this problem in the early 1980s and used it as an illustrative example to show that it was possible to send anonymous messages with unconditional sender and recipient untraceability.
  Anonymous communication networks based on this problem are often referred to as DC-nets (where DC stands for "dining cryptographers""")
    st.write("""Three cryptographers gather around a table for dinner. The waiter informs them that the meal has been paid for by someone, who could be one of the cryptographers or the National Security Agency (NSA). The cryptographers respect each other's right to make an anonymous payment, but want to find out whether the NSA paid. So they decide to execute a two-stage protocol""")
    st.image("images/crypto.png", width=500,)



    ab=st.selectbox('A and B: Head or Tail',['Head','Tail'])
    ac=st.selectbox('A and C: Head or Tail',['Head','Tail'])
    bc=st.selectbox('B and C: Head or Tail',['Head','Tail'])
    wp=st.selectbox('Who Paid',['A','B','C','NSA','A and B'])
    if wp=='A':
      vwp=[1,0,0]
    if wp=='B':
      vwp=[0,1,0]
    if wp=='C':
      vwp=[0,0,1]
    if wp=='NSA':
      vwp=[0,0,0]
    if wp=='A and B':
      vwp=[1,1,0]
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### ii - Diagram: ')

    test=dining_crypt(ab=ab,ac=ac,bc=bc,who_paid=vwp)
    st.graphviz_chart(test)

    #st.write(f"        ")
    #st.write(f"        ")
    #st.markdown('#### iv - Input Table: ')


    snippet = f"""
  def dining_crypt(ab="Head",ac="Tail",bc="Head",who_paid=[0,0,0]):
  dining=game(3,["A","B","C"])
  ab=ab
  ac=ac
  bc=bc
  list_value=[ac,ab,bc]
  #mat=[[0,ab,ac],[ab,0,bc],[ac,bc,0]]
  mat=[[0,ab,ac],[0,0,0],[0,bc,0]]
  dining.make_transition(mat)

  for id in range(len(list_value)):
  if list_value[id]=="Head":
      list_value[id]=1
  else:
      list_value[id]=0
  list_XOR=[]
  for kk in range(3):
  if not(who_paid[kk]):
      list_XOR.append(list_value[kk]^list_value[(kk+1)%3])
  else:
      list_XOR.append(not(list_value[kk]^list_value[(kk+1)%3]))

  list_XOR_label=[]
  for XOR in list_XOR:
  if XOR:
      list_XOR_label.append("True")
  else:
      list_XOR_label.append("False")
  rez=list_XOR[0]^list_XOR[1]^list_XOR[2]
  f=dining.display_diagram(list_XOR_label)
  if rez==0:
  if sum(who_paid)==0:
      f.attr(label='Result:NSA paid. Good Prediction',frontcolor='red')
  else:
      f.attr(label='Result:NSA paid. Bad Prediction')
  else:
  if sum(who_paid)==0:
   f.attr(label='Result: Someone paid. Bad Prediction')
  else:
      f.attr(label='Result: Someone paid. Good Prediction')

  return f

    """
    code_header_placeholder = st.empty()
    snippet_placeholder = st.empty()

    st.write(f"        ")
    st.write(f"        ")
    code_header_placeholder.markdown(f"#### iii - Code 02 - Dining Cryptographers")
    snippet_placeholder.code(snippet)
    st.write("     ")
    st.write("     ")
    st.markdown("---")



  ### 03
  if nlp_steps == "03 - ICGS":
    st.markdown("---")
    st.write(f"        ")

    st.header("03 - ICGS")
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### i - Quick Introduction: ')
    st.text('''
    Impartial combinatorial games (ICGs)

    Integration: We have to initialze three parameters:
    ActT1: T1's action, ="r" (right) or "l" (left)
    ActT2: T2's action, ="r" (right) or "s" (straight) or "l" (left)
    ActC: Controller's action, ="e" or "a" or "o" (see description below)
    The blue connection are the state that cannot be distinguish for the agent in the label.
    The red node is the final state after all the actions
    The iCGS describes a variant of the Train Gate Controller scenario (Alur, Henzinger, and Kupferman 2002).
    Two trains t1 and t2 pass through a crossroad. Due to agreements between the railway companies, train t1 can choose between the right (r) or left (l) track,
    while t2 can choose between the right (r), left (l) or straight (s) track.
    At the same time, controller c has to select the right combination of tracks. For example, if t1 and t2 choose the joint action r, then c has to select action 1 to proceed to the next step.
    Moreover, train t1 has partial observability on the choices of t2. For instance, if t1 chooses l, then she cannot distinguish whether t2 selects r or s, but she would observe if t2 chose l as well.
    After this first step, c can still change her mind. Specifically, she can change arbitrarily the selection of tracks (e), request a new choice to the trains (a), or execute their selection (o).
    The controller c has partial observability, she cannot distinguish between s2 and s3, i.e.
    she does not distinguish r and l of t1 when t2 selects l. Moreover, ∗ denotes any tuple of actions for which a transition is not given explicitly.''')


    st.write(f"        ")
    st.markdown('### ii Diagram Presentation')
    test=iCGS_train(boolean_Red=False)
    st.write(f"        ")
    st.graphviz_chart(test)
    st.write("       ")
    st.markdown('#### iii - Strategy')
    st.write("        ")
    st.write('     ')
    t1_act=st.selectbox('Action of t1 in the initial state',['right','left'])
    st.write('      ')
    t2_act=st.selectbox('Action of t2 in the initial state',['right','straight', 'left'])
    st.write('     ')
    c_act=st.selectbox('Action of controller in the state after',['6','5','4','3','2','1'])
    st.write('    ')
    cs1=st.selectbox('Action in s1',['a','o', 'e'])
    st.write('    ')
    cs2=st.selectbox('Action in s2',['a','o', 'e'])
    st.write('    ')
    cs3=st.selectbox('Action in s3',['o','a', 'e'])
    st.write('    ')
    cs4=st.selectbox('Action in s4',['o','a', 'e'])
    st.write('    ')
    cs5=st.selectbox('Action in s5',['e','a', 'o'])
    st.write('    ')
    cs6=st.selectbox('Action in s6',['e','a', 'o'])

    t1_act,t2_act=t1_act[0],t2_act[0]



    st.write(f"        ")
    st.write(f"        ")
    if st.button('Print Diagram'):
      st.markdown('#### iii - Diagram: ')
      test=iCGS_train(t1_act,t2_act,c_act,cs1,cs2,cs3,cs4,cs5,cs6)
      st.graphviz_chart(test)



    code_header_placeholder = st.empty()
    snippet_placeholder = st.empty()

    st.write(f"        ")
    st.write(f"        ")
    st.write("     ")
    st.markdown("---")



  ### 5
  if nlp_steps == '05 - Buchi  Automaton':
    st.markdown("---")
    st.write(f"        ")

    st.header('05 - Buchi  Automaton')
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### i - Quick Introduction: ')
    st.write(f"        ")
    st.write(f"        ")


    st.markdown("#### ii - Select Parameters: ")

    with open('data/test.txt', 'r') as f:
      s = f.read()
    input_ast_format = ast.literal_eval(s)


    input_initial_state = st.selectbox(
    'Select initial state:',
    list(input_ast_format.keys()))


    input_final_states = st.multiselect(
    'Select final states:',
    list(input_ast_format.keys()),list(np.random.choice(list(input_ast_format.keys()),2)))


    dfa = VisualDFA(
    states={"q0", "q1", "q2", "q3", "q4"},
    input_symbols={"0", "1"},
    transitions=input_ast_format,
    initial_state=input_initial_state,
    final_states=set(input_final_states),
    )


    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### iii - Diagram: ')
    digraph_output = dfa.show_diagram("101")
    st.graphviz_chart(digraph_output)

    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### iv - Input Table: ')
    digraph_table = dfa.input_check("10011")
    st.table(digraph_table)


    snippet = f"""
    dfa = VisualDFA(
    states={"q0", "q1", "q2", "q3", "q4"},
    input_symbols={"0", "1"},
    transitions=input_ast_format,
    initial_state=input_initial_state,
    final_states=set(input_final_states),
    )


    """
    code_header_placeholder = st.empty()
    snippet_placeholder = st.empty()

    st.write(f"        ")
    st.write(f"        ")
    code_header_placeholder.markdown(f"#### v - Code 04 - Buchi  Automaton")
    snippet_placeholder.code(snippet)
    st.write("     ")
    st.write("     ")
    st.markdown("---")



  ### 02
  if nlp_steps == '02 - Strategy Example':
    st.markdown("---")
    st.write(f"        ")

    st.header("02 - Strategy Example")
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### i - Quick Introduction: ')
    st.text(''' This is an example of a game involving very simple states, but with a strategy system.   ''')


    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### ii - Strategy')
    P10=st.selectbox('Action of P1 in the initial state',['A','B','Do nothing'])
    P11=st.selectbox('Action of P1 in the state 1',['A','B','Do nothing'])
    P12=st.selectbox('Action of P1 in the state 2',['A','B','Do nothing'])
    P13=st.selectbox('Action of P1 in the state 3',['A','B','Do nothing'])
    P20=st.selectbox('Action of P2 in the initial state',['C','D','Do nothing'])
    P21=st.selectbox('Action of P2 in the state 1',['C','D','Do nothing'])
    P22=st.selectbox('Action of P2 in the state 2',['C','D','Do nothing'])
    P23=st.selectbox('Action of P2 in the state 3',['C','D','Do nothing'])
    if P10=='Do nothing':
      P10='0'
    if P11=='Do nothing':
      P11='0'
    if P12=='Do nothing':
      P12='0'
    if P13=='Do nothing':
      P13='0'
    if P20=='Do nothing':
      P20='0'
    if P21=='Do nothing':
      P21='0'
    if P22=='Do nothing':
      P22='0'
    if P23=='Do nothing':
      P23='0'
    st.write(f"        ")
    st.write(f"        ")
    st.markdown('#### iii - Diagram: ')
    test=example_strategy(P10+P11+P12+P13,P20+P21+P22+P23)
    st.graphviz_chart(test)


    snippet = f"""


    """
    code_header_placeholder = st.empty()
    snippet_placeholder = st.empty()

    st.write(f"        ")
    st.write(f"        ")
    st.write("     ")
    st.write("     ")
    st.markdown("---")


  ### 6
  if nlp_steps == '06 - Upload Configuration':
    upload_file_handler()

def upload_file_handler():
  def file_select(folder='./data'):
      filelist=os.listdir(folder)
      st.markdown("OR")
      selectedfile=st.selectbox('Select a default dataset',filelist)
      return os.path.join(folder,selectedfile)


  st.markdown("---")
  st.write(f"        ")

  st.header('06 - Upload File')
  st.write(f"        ")
  st.write(f"        ")
  st.markdown("Upload Txt File with config")
  st.button('Upload Data')
  uploaded_file=st.file_uploader('Upload Dataset in .txt',type=['TXT'])
  s = None
  filename = None
  if uploaded_file is not None:
    # st.write('hello world')
    with open('data/' + str(uploaded_file.name), 'w') as f:
        # st.write(dir(uploaded_file))
        data = uploaded_file.getvalue().decode('utf-8').splitlines()
        # data = uploaded_file.getvalue().decode('utf-8').splitlines()
        # st.write(bytes_data)
        # st.write(data[1])
        # st.write(type(data))
        for line in data:
            f.write(str(line) + '\n')
  else:
    filename=file_select()
    st.info('You selected {}'.format(filename))
    with open(str(filename), 'r') as f:
      s = f.read()
  #
  # if filename:
  #   st.write(str(filename))
  #   game_strategy=game(load_file=True,path1=str(filename),AW=True)
  #   # st.write(game_strategy.name_list)
  #   graph=game_strategy.display_diagram()
  #   st.write(str(graph))
  #   st.graphviz_chart(graph)

  st.write("     ")
  st.markdown("Example of config file")
  if s is not None:
    snippet = s
  else:
    snippet = f"""
  Transition
  0 AC,AD BC,BD 0
  0 0 AD,BD AC,BC
  0 0 0 AD,BC
  0 0 0 *
  Name_State
  s0 s1 s2 s3
  Initial_State
  s0
  peppia

  """
  snippet_placeholder = st.empty()

  st.write(f"        ")
  st.write(f"        ")
  snippet_placeholder.code(snippet)
  st.write("     ")
  st.markdown("---")

  return filename



"""def D_parser_test():
  st.header("Parser")
  st.write(f"    ")
  print("logic")
  print(st.session_state.info_model_test[0][0])
  if st.session_state.info_model_test[0][0]=='LTL':
    print("formula")
    print(st.session_state.info_model_test[0][1])
    lexer = LTLLexer(st.session_state.info_model_test[0][1])
    print("lexer")
    print(lexer)
    stream = CommonTokenStream(lexer)
    print("stream")
    print(stream)
    parser = LTLParser(stream)
    print("parser")
    print(parser)
    tree = parser.ltlExpr()
    visitor = YOUR_VISITOR_CLASS()
    print(visitor.visit(tree))"""

def upload_xml_file_handler():
  def file_select(folder='./xml_data'):
      filelist=os.listdir(folder)
      st.markdown("OR")
      selectedfile=st.selectbox('Select a default dataset',filelist)
      return os.path.join(folder,selectedfile)


  st.markdown("---")
  st.write(f"        ")

  st.header('06 - Upload File')
  st.write(f"        ")
  st.write(f"        ")
  st.markdown("Upload XML File with config")
  st.button('Upload Data')
  uploaded_file=st.file_uploader('Upload Dataset in .xml',type=['XML'])
  s = None
  filename = None
  if uploaded_file is not None:
    # st.write('hello world')
    with open('xml_data/' + str(uploaded_file.name), 'w') as f:
        # st.write(dir(uploaded_file))
        data = uploaded_file.getvalue().decode('utf-8').splitlines()
        # data = uploaded_file.getvalue().decode('utf-8').splitlines()
        # st.write(bytes_data)
        # st.write(data[1])
        # st.write(type(data))
        for line in data:
            f.write(str(line) + '\n')
  else:
    filename=file_select()
    st.info('You selected {}'.format(filename))
    with open(str(filename), 'r') as f:
      s = f.read()
  return filename

def parser(formula):
    change=[[' ',''],['||','|'],['&&','&'],['or', '|'],['and','&'],['implies','>'],['->','>'],['next','X'],['eventually','F'],['always','G']]
    operatorSimpleL=[')',']']
    operatorSimpleR=['X','F','G','(','[']
    operatorDouble=['|','&','U','R','>']
    closeoperator=[')',']']
    openoperator=['(','[']
    operator=operatorSimpleL+operatorSimpleR+operatorDouble
    for operator_change in change:
        formula=formula.replace(operator_change[0],operator_change[1])
    def cut(formula):
        new_formula=[]
        tmp=''
        for letter in formula:
            if letter in operator:
                if tmp!='':
                    new_formula.append(tmp)
                new_formula.append(letter)
                tmp=''
            else:
                tmp+=letter
        new_formula.append(tmp)
        return new_formula
    cut_formula=cut(formula)
    def verif_paranthese(formula):
        cmpt=0
        for obj in formula:
            if obj in openoperator:
                cmpt+=1
            elif obj in closeoperator:
                cmpt+=-1
        return not(cmpt)
    if not(verif_paranthese(cut_formula)):
        return False

    cut_class=[]
    for obj in cut_formula:
        if obj in operatorDouble:
            cut_class.append('D')
        elif obj in operatorSimpleL:
            cut_class.append('L')
        elif obj in operatorSimpleR:
            cut_class.append('R')
        else:
            cut_class.append('AP')
    def verif(list_cut):
        if not(list_cut[0]=='AP' or list_cut[0]=='L'):
            return False
        for id_obj in range(len(list_cut)-1):
            print(id_obj)
            if list_cut[id_obj]=='R':
                if (list_cut[id_obj+1]=='L' or list_cut[id_obj+1]=='D'):
                    return False
            elif list_cut[id_obj]=='D':
                if (list_cut[id_obj+1]=='L' or list_cut[id_obj+1]=='D'):
                    return False
            elif list_cut[id_obj]=='AP':
                if (list_cut[id_obj+1]=='R' or list_cut[id_obj+1]=='AP') :
                    return False
            elif list_cut[id_obj]=='L':
                if (list_cut[id_obj+1]=='AP' or list_cut[id_obj+1]=='L' or list_cut[id_obj+1]=='R'):
                    return False
        if (list_cut[-1]=='R' or list_cut[-1]=='D'):
            return False
        return True
    return verif(cut_class),cut_formula,cut_class





def display_MCMAS():
  if st.session_state.cmpt_model==-1:
    st.markdown("---")
    st.write(f"    ")
    st.header("Model Checking for MAS")
    st.write(f"    ")
    st.write(f"    ")
    st.markdown('Logic Selection ')
    Logic=st.selectbox('Select your logic',['ATL','CTL','LTL','SL','CapATL', 'OL', 'OATL', 'RBATL', 'RABATL', 'NatATL'])
    st.write(f"    ")
    st.write(f"    ")
    formula=st.text_input('Write your formula',' ')
    st.write("     ")
    st.write('Your formula with the '+Logic+' logic is '+formula)
    st.write("     ")
    Verif,list_pars,list_type=parser(formula)
    st.write(str(Verif))
    st.write(str(list_pars))
    st.write("     ")
    st.markdown("---")
    if st.button('Next : To Parser'):
      (st.session_state.info_model_test).append([Logic,formula])
      st.session_state.cmpt_model=-2
      st.experimental_rerun()
  if st.session_state.cmpt_model==-2:
    pass
    #D_parser_test()

def read_markdown_file(markdown_file):
  return Path(markdown_file).read_text()

def display_MS(page):
  if page == 0:
    intro_markdown = read_markdown_file("./README.md") 
    st.markdown(intro_markdown, unsafe_allow_html=True)
  elif page == 3:
    if st.session_state.cmpt_model<=0:
      D_agent()
    elif st.session_state.cmpt_model==1:
      D_state()
    elif st.session_state.cmpt_model==2:
      D_atoms()
    elif st.session_state.cmpt_model==3:
      D_state_atoms()
    elif st.session_state.cmpt_model==4:
      D_resource()
    elif st.session_state.cmpt_model==5:
      D_action()
    elif st.session_state.cmpt_model==6:
      # st.session_state.mat_transi=[]
      Na=len(st.session_state.info_model[0][0])
      D_transition(Na-st.session_state.info_model[0][1]-1,Na)
    elif st.session_state.cmpt_model==7:
      # st.session_state.mat_transi=[]
      st.session_state.info_model[0][1]=len(st.session_state.info_model[0][0])
      D_cost()
    elif st.session_state.cmpt_model==8:
      # st.session_state.mat_transi=[]
      st.session_state.info_model[0][1]=len(st.session_state.info_model[0][0])
      D_printgraph()
    # elif st.session_state.cmpt_model==5:
    #   Na=len(st.session_state.info_model[0][0])
    #   D_strategy(Na-st.session_state.info_model[0][1]-1,Na)
    elif st.session_state.cmpt_model==9:
      D_logic()
    elif st.session_state.cmpt_model==10:
      if st.session_state.info_model[9][0] == 'RBATL':
        from vitamin_model_checker.model_checker_interface.explicit.RBATL import RBATL
        result = RBATL.model_checking(st.session_state.info_model[9][1], 'data/tmp.txt')
        del RBATL
        st.write(result['res'])
        st.write(result['initial_state'])
      else:
        from vitamin_model_checker.model_checker_interface.explicit.ATL import ATL
        result = ATL.model_checking(st.session_state.info_model[9][1], 'data/tmp.txt')
        del ATL
        st.write(result['res'])
        st.write(result['initial_state'])
      
      #D_parser()

    st.markdown("---")
    st.write(f"    ")
    st.header("Model design for User")
    st.write(f"    ")
    st.markdown("---")
  elif page==4:
    filename = upload_file_handler()
    st.markdown('Logic Selection ')
    Logic=st.selectbox('Select your logic',['ATL','ATLF','CTL','LTL','SL','CapATL', 'OL', 'OATL', 'NatATL', 'RBATL', 'RABATL'])
    st.write(f"    ")
    st.write(f"    ")
    formula=st.text_input('Write your formula',' ')
    st.write("     ")
    #st.write('Your formula with the '+Logic+' logic is '+formula)
    #st.write("     ")
    """Verif,list_pars,list_type=parser(formula)
    st.write(str(Verif))
    st.write(str(list_pars))
    st.write(str(list_type))"""

    # read_input.read_file(filename) 
    # res_parsing = do_parsing(formula, read_input.get_number_of_agents())

    # if formula == '':
    #   st.write("please enter a formula")
    # st.write("Valid formula: " +str(res_parsing))
    # st.write('Selected file: ' + str(filename))
    # with open(filename) as f:
    #     # st.write('Content:')
    #     content = f.read()
    #     # st.write(content)
    start_time = None
    if st.button('Next : To Model Checking'):
      start_time = time.time()
      if Logic == 'ATL':
        from vitamin_model_checker.model_checker_interface.explicit.ATL import ATL
        result = ATL.model_checking(formula, filename)
        del ATL
        st.write(result['res'])
        st.write(result['initial_state'])
      elif Logic == 'CapATL':
        from vitamin_model_checker.model_checker_interface.explicit.CapATL import CapATL
        result = CapATL.model_checking(formula, filename)
        del CapATL
        st.write(result)
        # st.write(result['initial_state'])
      elif Logic == 'ATLF':
        from vitamin_model_checker.model_checker_interface.explicit.ATLF import ATLF
        result  = ATLF.model_checking(formula, filename)
        del ATLF
        st.write(result['res'])
        st.write(result['initial_state'])
      elif Logic == 'OL':
        from vitamin_model_checker.model_checker_interface.explicit.OL import OL
        result  = OL.model_checking(formula, filename)
        del OL
        st.write(result['res'])
      elif Logic == 'OATL':
        from vitamin_model_checker.model_checker_interface.explicit.OATL import OATL
        result  = OATL.model_checking(formula, filename)
        del OATL
        st.write(result['res'])
        st.write(result['initial_state'])
      elif Logic == 'RBATL':
        from vitamin_model_checker.model_checker_interface.explicit.RBATL import RBATL
        result  = RBATL.model_checking(formula, filename)
        del RBATL
        st.write(result['res'])
        st.write(result['initial_state'])
      elif Logic == 'RABATL':
        from vitamin_model_checker.model_checker_interface.explicit.RABATL import RABATL
        result  = RABATL.model_checking(formula, filename)
        del RABATL
        st.write(result['res'])
        st.write(result['initial_state'])
      elif Logic == 'NatATL':
        from vitamin_model_checker.model_checker_interface.explicit.NatATL import NatATL
        result = NatATL.model_checking(formula, filename)
        del NatATL
        st.write(result)
      elif Logic == 'CTL':
        from vitamin_model_checker.model_checker_interface.explicit.CTL import CTL
        result = CTL.model_checking(formula, filename)
        del CTL
        st.write(result)
      elapsed_time = time.time() - start_time
      st.write("Execution time:", format_time(elapsed_time))
      start_time = None
      # formula - we have the string literal of the input formula
      # filename - the path to the model file
      # content - string denoting the content of the file
      # model_checking(formula, content)
      (st.session_state.info_model).append([Logic,formula])
      st.session_state.cmpt_model=7
      #st.experimental_rerun()
  else:
      xml = upload_xml_file_handler()
      
      def mapAttackGraphToCGS(xml):
        writting(xml, 'xml_data/cgs' './xml_data/cond.txt', ['Cond 4'])
        
        return cgs, '<Jn>G!goal'
        
      
      # Clement part to produce xml
      #      xml --> cgs
      #          --> formula
      cgs,formula = mapAttackGraphToCGS(xml)
      if st.button('Next : To Model Checking'):
        from vitamin_model_checker.model_checker_interface.explicit.ATL import ATL
        result = ATL.model_checking(formula, cgs)
        del ATL
        st.write(result['res'])
        st.write(result['initial_state'])


def D_agent():
  st.write(f"    ")
  st.write(f"    ")
  Na=st.selectbox('Number of agent',[str(i) for i in range(1, 15)])
  st.write(f"    ")
  List_name_agent=[]
  for id_agent in range(int(Na)):
    tmp=st.text_input('Name of the agent number '+str(id_agent),'A'+str(id_agent))
    List_name_agent.append(tmp)
  st.write(f"    ")
  if st.button('Next : To State'):
    (st.session_state.info_model).append([List_name_agent,len(List_name_agent)-1])
    st.session_state.cmpt_model=1
    print(st.session_state.info_model[0])
    st.experimental_rerun()


def D_state():
  NS=st.selectbox('Number of State',[str(i) for i in range(1, 25)])
  st.write("     ")
  List_name_state=[]
  for id_state in range(int(NS)):
    tmp=st.text_input('Name of the state number '+str(id_state),'s'+str(id_state))
    List_name_state.append(tmp)
  st.write(f"     ")
  if st.button('Next : To Atoms'):
    (st.session_state.info_model).append(List_name_state)
    st.session_state.cmpt_model=2
    st.experimental_rerun()

def D_atoms():
  NS=st.selectbox('Number of Atoms',[str(i) for i in range(1, 25)])
  st.write("     ")
  List_name_atoms=[]
  for id_state in range(int(NS)):
    tmp=st.text_input('Name of the atom '+str(id_state),'ap'+str(id_state))
    List_name_atoms.append(tmp)
  st.write(f"     ")
  if st.button('Next : To State-Atoms'):
    st.session_state.info_model.append(List_name_atoms)
    st.session_state.cmpt_model=3
    st.experimental_rerun()

def D_state_atoms():
  st.write('For each state')
  st.write("     ")
  List_name_state=st.session_state.info_model[1]
  NS=len(List_name_state)
  List_atoms=st.session_state.info_model[2]
  atoms = []
  for id_state in range(NS):
    tmp=st.multiselect(f'Labels of state {List_name_state[id_state]}', List_atoms,List_atoms[0])
    atoms.append((tmp))
  if st.button('Next : To Resources'):
    st.session_state.info_model.append(atoms)
    st.session_state.cmpt_model=4
    st.experimental_rerun()

def D_resource():
  NS=st.selectbox('Number of Resources',[str(i) for i in range(0, 11)])
  st.write("     ")
  List_name_res=[]
  for id_state in range(int(NS)):
    tmp=st.text_input('Name of the resource '+str(id_state),'r'+str(id_state))
    List_name_res.append(tmp)
  st.write(f"     ")
  if st.button('Next : To Actions'):
    (st.session_state.info_model).append(List_name_res)
    st.session_state.cmpt_model=5
    st.experimental_rerun()

def D_action():
  Nact=st.selectbox('Number of Action',[str(i) for i in range(1, 25)])
  Alphabet=list(string.ascii_uppercase)
  st.write("     ")
  List_name_action=['*','No Action']
  for id_action in range(int(Nact)):
    tmp=st.text_input('Name of the action number '+str(id_action),Alphabet[id_action])
    List_name_action.append(tmp)
  st.write(f"     ")
  if st.button('Next : To Transitions'):
    (st.session_state.info_model).append(List_name_action)
    st.session_state.cmpt_model=6
    st.experimental_rerun()


def D_transition(act_input,Na):
  st.write('Design your graph')
  st.write('For each couple of state')
  st.write("     ")
  Mat_transition=[]
  List_name_state=st.session_state.info_model[1]
  NS=len(List_name_state)
  List_action=st.session_state.info_model[5]
  for id_state1 in range(NS):
    List_transition=[]
    for id_state2 in range(int(NS)):
      tmp=st.multiselect('Action of '+st.session_state.info_model[0][0][act_input]+ ' in the state '+List_name_state[id_state1]+' to '+List_name_state[id_state2]+'.',List_action,List_action[1])
      List_transition.append(tmp)
    Mat_transition.append(List_transition)
  st.write("     ")
  if (Na-act_input)>1:
    if st.button('Next : To '+st.session_state.info_model[0][0][act_input+1]):
      st.session_state.info_model[0][1]+=(-1)
      (st.session_state.mat_transi).append(Mat_transition)
      st.session_state.cmpt_model=6
      st.experimental_rerun()
  else:
    if st.session_state.info_model[4]:
      if st.button('Next : To Cost'):
        (st.session_state.mat_transi).append(Mat_transition)
        st.session_state.cmpt_model=7
        Mat_to_Label()
        st.experimental_rerun()
    else:
      if st.button('Next : To Graph'):
        (st.session_state.mat_transi).append(Mat_transition)
        st.session_state.cmpt_model=8
        Mat_to_Label()
        st.session_state.info_model.append([])
        st.experimental_rerun()

def D_cost():
  st.write('Add the costs for the actions')
  st.write("     ")
  Costs=[]
  for i in range(0, len(st.session_state.info_model[6])):
     from_state = st.session_state.info_model[1][i]
     for j in range(0, len(st.session_state.info_model[6][i])):
        to_state = st.session_state.info_model[1][j]
        if 'No Action' not in st.session_state.info_model[6][i][j]:
          res = (from_state, st.session_state.info_model[6][i][j], [])
          for k in range(0, len(st.session_state.info_model[4])):
             tmp=st.selectbox(f'Cost of joint action {st.session_state.info_model[6][i][j]} in {from_state} to state {to_state} for resource {st.session_state.info_model[4][k]}', range(0, 11))
             res[2].append(tmp)
          Costs.append(res)
  # NS=len(List_name_state)
  # List_action=st.session_state.info_model[2]
  # for id_state1 in range(NS):
  #   List_transition=[]
  #   for id_state2 in range(int(NS)):
  #     tmp=st.multiselect('Action of '+st.session_state.info_model[0][0][act_input]+ ' in the state '+List_name_state[id_state1]+' to '+List_name_state[id_state2]+'.',List_action,List_action[1])
  #     List_transition.append(tmp)
  #   Mat_transition.append(List_transition)
  # st.write("     ")
  if st.button('Next : To Graph'):
    st.session_state.info_model.append(Costs)
    st.session_state.cmpt_model=8
    # Mat_to_Label()
    st.experimental_rerun()

def Mat_to_Label():
  def concat(list_str):
    tmp1=''
    for stri in list_str:
      tmp1+=stri
    return tmp1

  print(st.session_state.mat_transi)
  Mat=[]
  for s_in in range(len(st.session_state.info_model[1])):
    Lis=[]
    for s_out in range(len(st.session_state.info_model[1])):
      tmp=''
      for mat_label in st.session_state.mat_transi:
        tmp+=concat(mat_label[s_in][s_out])
        tmp+='|'
      Lis.append(tmp[:-1])
    Mat.append(Lis)
  print(Mat)
  st.session_state.info_model.append(Mat)



def D_printgraph():
  st.write('Graph')
  st.write("     ")
  st.markdown('#### iii - Diagram: ')
  test=display_graph_MS(st.session_state.info_model[6],st.session_state.info_model[0][0],st.session_state.info_model[1])
  st.graphviz_chart(test)
  
  store(st.session_state.info_model[0][0], st.session_state.info_model[1], st.session_state.info_model[2], st.session_state.info_model[3], st.session_state.info_model[7], st.session_state.info_model[5], st.session_state.info_model[6], 'data/tmp.txt')
  
  if st.button('Next : To Logic'):
    (st.session_state.info_model).append(test)
    st.session_state.cmpt_model=9
    st.experimental_rerun()

def store(agents, states, atoms, labelling, costs, actions, transitions, path):
   with open(path, 'w') as file:
      no_actions = '|'.join(['No Action' for i in range(0, len(agents))])
      revised_transitions = ''
      for tr in transitions:
        for act in tr:
           if act == no_actions:
              revised_transitions += '0 '
           else:
              revised_transitions += act.replace('|', '') + ' '
        revised_transitions += '\n'
      revised_transitions = revised_transitions[:-1]
      labels = ''
      for state_labels in labelling:
        for atom in atoms:
          if atom in state_labels:
            labels += '1 '
          else:
            labels += '0 '
        labels += '\n'
      labels = labels[:-1]
      # labels = ''
      # for i in range(0, len(states)):
      #   for j in range(0, len(states)):
      #     if i == j:
      #       labels += '1 '
      #     else:
      #       labels += '0 '
      #   labels += '\n'
      # labels = labels[:-1]
      l_agents= len(agents)
      file.write(f'''
Transition
{revised_transitions}
Name_State
{' '.join(states)}
Initial_State
s0
Atomic_propositions
{' '.join(atoms)}
Labelling
{labels}
Number_of_agents
{l_agents}
''')
      if costs:
         file.write('Costs_for_actions\n')
         for c in costs:
            file.write('{a} {s}${res}'.format(a=c[1].replace('|', ''), s=c[0], res=':'.join([str(v) for v in c[2]])))
            file.write('\n')

def D_logic():
  # st.header("Model Checking for MAS")
  st.write(f"    ")
  st.write(f"    ")
  st.markdown('Logic Selection ')
  Logic=st.selectbox('Select your logic',['ATL','RBATL'])

  st.write(f"    ")
  st.write(f"    ")
  formula=st.text_input('Write your formula',' ')
  st.write("     ")
  st.write('Your formula with the '+Logic+' logic is '+formula)
  st.write("     ")
  Verif,list_pars,list_type=parser(formula)
  st.write(str(Verif))
  st.write(str(list_pars))
  st.write(str(list_type))
  if st.button('Next : To Model Checking'):
    (st.session_state.info_model).append([Logic,formula])
    st.session_state.cmpt_model=10
    st.experimental_rerun()


#def D_parser():
#  st.header("Parser")
#  st.write(f"    ")
#  if st.session_state.info_model[5][0]=='LTL':
#    lexer = LTLLexer(st.session_state.info_model[5][1])
#    stream = CommonTokenStream(lexer)
#    parser = LTLParser(stream)
#    tree = parser.ltlExpr()
#    visitor = YOUR_VISITOR_CLASS()
#    print(visitor.visit(tree))


def D_strategy(act_input,id_state1):
  st.write('Design your strategy')
  st.write('For each agent in each state')
  st.write("     ")
  List_strategy=[]
  List_name_state=st.session_state.info_model[1]
  NS=len(List_name_state)
  List_action=st.session_state.info_model[2]
  for id_state1 in range(NS):
    tmp=st.selectbox('Action of '+st.session_state.info_model[0][0][act_input]+ ' in the state '+List_name_state[id_state1]+'.',List_action)
    print(tmp)
    List_strategy.append(tmp)
  st.write("     ")
  if (id_state1-act_input)>1:
    if st.button('Next : To '+st.session_state.info_model[0][0][act_input+1]):
      st.session_state.info_model[0][1]+=(-1)
      (st.session_state.mat_transi).append(List_strategy)
      st.session_state.cmpt_model=6
      st.experimental_rerun()
  else:
    if st.button('Next : To Logic'):
      (st.session_state.mat_transi).append(List_strategy)
      st.session_state.cmpt_model=7
      st.session_state.info_model.append(st.session_state.mat_transi)
      print(st.session_state.info_model)
      st.experimental_rerun()



# timer (time elapsed)
def format_time(seconds):
  milliseconds = int(seconds * 1000) % 1000
  return f"{int(seconds):02d}.{milliseconds:03d}"





import numpy as np
from vitamin_model_checker.models.CGS.CGS import CGS

class costCGS(CGS):
    def init(self):
        super(costCGS, self).init()
        self.costs = []
        self.cost_for_action = {}
        self.usesCostsInsteadOfActions = False


    def read_file(self, filename):
        with open(filename, 'r') as f:
            lines = f.readlines()

        self.graph = []
        self.states = []
        self.atomic_propositions = []
        self.matrix_prop = []
        self.initial_state = ''
        self.number_of_agents = ''
        self.cost_for_action = {}
        self.usesCostsInsteadOfActions = False

        current_section = None
        transition_content = ''
        unknown_transition_content = ''
        name_state_content = ''
        atomic_propositions_content = ''
        labelling_content = ''
        rows_graph = []
        rows_prop = []
        

        for line in lines:
            line = line.strip()
            #Section "header"
            if line == 'Transition':
                current_section = 'Transition'
            elif line == 'Transition_With_Costs':
                current_section = 'Transition'
                self.usesCostsInsteadOfActions = True
            elif line == 'Unkown_Transition_by':
                current_section = 'Unknown_Transition_by'
            elif line == 'Name_State':
                current_section = 'Name_State'
            elif line == 'Initial_State':
                current_section = 'Initial_State'
            elif line == 'Atomic_propositions':
                current_section = 'Atomic_propositions'
            elif line == 'Labelling':
                current_section = 'Labelling'
            elif line == 'Number_of_agents':
                current_section = 'Number_of_agents'
            elif line == 'Costs_for_actions':
                current_section = 'Costs_for_actions'
            elif line == 'Costs_for_actions_split':
                current_section = 'Costs_for_actions_split'

            #If not header, then read contents based on what section we are in
            
            elif current_section == 'Transition':
                transition_content += line + '\n'
                values = line.strip().split()
                rows_graph.append(values)
            
            elif current_section == 'Unknown_Transition_by':
                unknown_transition_content += line + '\n'
            elif current_section == 'Name_State':
                name_state_content += line + '\n'
                values = line.strip().split()
                self.states = np.array(values)
            elif current_section == 'Initial_State':
                self.initial_state = line
            elif current_section == 'Atomic_propositions':
                atomic_propositions_content += line + '\n'
                values = line.strip().split()
                self.atomic_propositions = np.array(values)
            elif current_section == 'Labelling':
                labelling_content += line + '\n'
                values = line.strip().split()
                rows_prop.append(values)
            elif current_section == 'Number_of_agents':
                self.number_of_agents = line
            elif current_section == "Costs_for_actions":
                values = line.strip().split()
                action_name = values[0]
                state_and_cost_string = values[1].split(";")
                for couple in state_and_cost_string:
                    state_and_cost = couple.split("$")
                    costs = [int(c) for c in state_and_cost[1].split(':')]
                    self.cost_for_action.update({self.translate_action_and_state_to_key(action_name, state_and_cost[0]): costs})
            elif current_section == "Costs_for_actions_split":
                values = line.strip().split()
                action_name = values[0]
                state_and_cost_string = values[1].split(";")
                for couple in state_and_cost_string:
                    state_and_cost = couple.split("$")
                    costs_res = state_and_cost[1].split(':')
                    costs = [[int(cc) for cc in c.split(',')] for c in costs_res]
                    self.cost_for_action.update({self.translate_action_and_state_to_key(action_name, state_and_cost[0]): costs})
        
        actions =[]
        a = 0
        grafo_prov = np.array(rows_graph)
        for row in grafo_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                else:
                    if (self.usesCostsInsteadOfActions):
                        new_row.append(item)
                    else:
                        new_row.append(str(item))
                        a = item.split(",")
                        for elem in a :
                            actions.append(elem)
            self.graph.append(new_row)   

        matrix_prop_prov = np.array(rows_prop)
        for row in matrix_prop_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                elif item == '1':
                    new_row.append(1)
                else:
                    new_row.append(str(item))
            self.matrix_prop.append(new_row)

    def read_from_model_object(self, model):
        super(costCGS, self).read_from_model_object(model)
        self.capacities_assignment = model.capacities_assignment
        self.action_capacities = model.actions_for_capacities
        self.capacities = np.array(model.capacities)
        self.cost_for_action = model.cost_for_action
        
    def get_cost_for_action(self, action, state):
        return self.cost_for_action[self.translate_action_and_state_to_key(action, state)]
        
    def get_cost_for_action_all(self):
        return self.cost_for_action


import re
import ply.lex as lex
import ply.yacc as yacc
import time

MAX_COALITION = 3

# Token
tokens = (
    'LPAREN',
    'RPAREN',
    'AND',
    'NOT',
    'IS',
    'UNTIL',
    'RELEASE',
    'NEXT',
    'KCAP',
    'PROP',
    'COALITION',
    'AGENT',
    # 'CAPACITY',
)


# Regular expressions for tokens
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_AND = r'&&|\&|and'
t_NOT = r'!|not'
t_IS = r'is|IS'
t_UNTIL = r'U|until'
t_RELEASE = r'R|release'
t_NEXT = r'X|next'
t_KCAP = r'K|kcap'
t_COALITION = r'<\d+(?:,\d+)*>'

t_AGENT = r'[0-9]+' 
t_PROP = r'[a-z]+'



def t_error(t):
    t.lexer.skip(1)



def p_expression_binary(p):
    '''expression : expression AND expression'''
    p[0] = (p[2], p[1], p[3])

def p_expression_not(p):
    '''expression : NOT expression'''
    p[0] = (p[1], p[2])

class CoalitionValueError(Exception):
    pass


def p_expression_ternary(p):
    '''expression : COALITION expression UNTIL expression 
                    | COALITION expression RELEASE expression'''
    coalition_values = re.findall(r'\d+', p[1])
    for value in coalition_values:
        if int(value) > int(MAX_COALITION):
            raise CoalitionValueError(f"Coalition value {value} exceeds maximum of {MAX_COALITION}")
    p[0] = (p[1] + p[3], p[2], p[4])


def p_expression_unary(p):
    '''expression :  COALITION NEXT expression'''
    p[0] = (p[1] + p[2], p[3])





def p_expression_group(p):
    '''expression : LPAREN expression RPAREN'''
    p[0] = p[2]

def p_expression_prop(p) :
    '''expression : PROP'''
    p[0] = p[1]


def p_expression_kcap(p):
    '''expression : KCAP AGENT expression2'''
    p[0] = (p[1] + p[2], p[3])


def p_expression_capformula_group(p):
    '''expression2 : LPAREN expression2 RPAREN'''
    p[0] = p[2]


def p_expression_capformula_and(p):
    '''expression2 : expression2 AND expression2'''
    p[0] = (p[2], p[1], p[3])


def p_expression_capformula_not(p):
    '''expression2 : NOT expression2 '''
    p[0] = (p[1], p[2])


def p_expression_capformula_is(p):
    '''expression2 : AGENT IS PROP'''
    p[0] = ( p[1] + p[3])


def p_error(p):
    pass


# lexer and parser
lexer = lex.lex()
parser = yacc.yacc()


def get_lexer():
    return lexer


# given an ATL formula as input and the max number of agents in the model,
# returns a tuple representing the formula divided into subformulas.
def do_parsing(formula, n_agent):
    global MAX_COALITION
    MAX_COALITION = n_agent
    try:
        result = parser.parse(formula)
        print(result)
        return result
    except SyntaxError:  # if parser fails
        print("syntax error")
        return None
    # except CoalitionValueError:  # coalition not existent
    #   print("pb of coalition")
    #   return None


# checks whether the input operator corresponds to a given operator defined in the grammar
def verify(token_name, string):
    lexer.input(string)
    for token in lexer:
        if token.type == token_name and token.value in string:
            return True
    return False





from vitamin_model_checker.models.costCGS.costCGS import *
from binarytree import Node
from vitamin_model_checker.logics.OL import *
from timeit import default_timer as timer

last_filename = ""
pre_set_array = []

# returns the states where the proposition holds
def get_states_prop_holds(prop):
    states = set()
    prop_matrix = cCGS.get_matrix_proposition()

    index = cCGS.get_atom_index(prop)
    if index is None:
        return None
    for state, source in enumerate(prop_matrix):
        if source[int(index)] == 1:
            states.add(state)
    return states



# set of states (ex. s1, s2) as input and returns a set of indices to identify them
def convert_state_set(state_set):
    states = set()
    for elem in state_set:
        position = cCGS.get_index_by_state_name(elem)
        states.add(int(position))
    return states
    
def convert_indices_state_set(indices_state_set):
    states = set()
    for elem in indices_state_set:
        states.add(cCGS.get_state_name_by_index(elem))
    return states


# converts a string into a set
def string_to_set(string):
    if string == 'set()':
        return set()
    set_list = string.strip("{}").split(", ")
    new_string = "{" + ", ".join(set_list) + "}"
    return eval(new_string)


#  function that builds a formula tree, used by the model checker
def build_tree(tpl):
    if isinstance(tpl, tuple):
        root = Node(tpl[0])
        if len(tpl) > 1:
            left_child = build_tree(tpl[1])
            if left_child is None:
                return None
            root.left = left_child
            if len(tpl) > 2:
                right_child = build_tree(tpl[2])
                if right_child is None:
                    return None
                root.right = right_child
    else:
        states = set()
        if (verify('FALSE', str(tpl))):
            return Node(str(states))
        elif (verify('TRUE', str(tpl))):
            return Node(str(set(cCGS.get_states())))
        states_proposition = get_states_prop_holds(str(tpl))
        if states_proposition is None:
            return None
        else:
            for element in states_proposition:
                states.add(cCGS.get_state_name_by_index(element))
            root = Node(str(states))
    return root

def complement(state_set):
    result = set()
    graph = cCGS.get_graph()
    if (len(graph) - len(state_set) <= 0): 
        return result
    for i in range(0, len(graph)):
        if (i not in state_set):
            result.add(i)
    return result

	
# pre with standard matrix
def pre(state_set):
    result = set()
    graph = cCGS.get_graph()
    for i, source in enumerate(graph):  # take states that have at least one transition to one of the states in the set
        for j in state_set:
            if graph[i][j] != 0:
                    result.add(i)
                    break
    return result

# builds an array where the j-th element is a set with all the predecessors of the j-th state
def build_pre_set_array():
    graph = cCGS.get_graph()
    graph_len = len(graph)
    result = [set()] * graph_len
    for i in range(0, graph_len):
        for j in range(0, graph_len):
            if graph[i][j] != 0:
                result[j].add(i)
    pre_set_array = result
               
# pre that uses the array of predecessors instead of the graph matrix
def pre_set_variant(state_set):
    result = set()
    for i in state_set:
        for j in pre_set_array[i]:
            result.add(i)
    return result

# triangle "right" operator 
def triangle(s, n, state_set):
    cost = 0
    tCost = 0
    graph = cCGS.get_graph()
    isLoopPresent = False
    for r in state_set:
        if (graph[s][r] != "*"):
            cost += int(graph[s][r])
        else:
            isLoopPresent = True
    # return ((cost <= n and (cost != 0 or isLoopPresent)))
    return (cost <= n)
    
# triangle "down" operator
def triangle_down(n, state_set):
    result = set()
    state_set = convert_state_set(state_set)
    state_set_complement = complement(state_set)
    predecessors = pre(state_set)
    for s in predecessors:
        if (triangle(s, n, state_set_complement)):
            result.add(s)
    return convert_indices_state_set(result)
    
    
# triangle "down" operator
def triangle_down_variant(n, state_set):
    result = set()
    state_set = convert_state_set(state_set)
    state_set_complement = complement(state_set)
    predecessors = pre_set_variant(state_set)
    for s in predecessors:
        if (triangle(s, n, state_set_complement)):
            result.add(s)
    return convert_indices_state_set(result)
    
    
# function that solves the formula tree. The result is the model checking result.
# It solves every node depending on the operator.
def solve_tree(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)
    i = 0
    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            all_states = set(cCGS.get_states())
            ris = all_states - states
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('GLOBALLY', node.value):  # e.g. <Jn>Gφ
            #<Jn>falseRφ
            n = int(node.value[2:-2])
            states1 = set()
            states2 = string_to_set(node.left.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down(n, p))
            node.value = str(p)

        elif verify('DEMONIC', node.value) and verify('NEXT', node.value):  # e.g. <Jn>Xφ
            n = int(node.value[2:-2])
            states = string_to_set(node.left.value)
            ris = triangle_down(n, states)
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('EVENTUALLY', node.value):  # e.g. <Jn>Fφ
            #<Jn>trueUϕ.
            n = int(node.value[2:-2])
            states1 = set(cCGS.get_states())
            states2 = string_to_set(node.left.value)
            p = set()
            t = states2
            while t != p:
                p = t
                t = states2 | (states1 & triangle_down(n, p))
            node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('UNTIL', node.value):  # e.g. <Jn>φUθ
            n = int(node.value[2:-2])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set()
            t = states2
            while t != p: 
                p = t
                t = states2 | (states1 & triangle_down(n, p))
            node.value = str(p)
        elif verify('DEMONIC', node.value) and verify('RELEASE', node.value): #e.g. <Jn>φRθ
            n = int(node.value[2:-2])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down(n, p))
            node.value = str(p)
        elif verify('DEMONIC', node.value) and verify('WEAK', node.value): #e.g. <Jn>φWθ
            #<Jn>(θ R (φ ∨ θ))
            n = int(node.value[2:-2])
            states1 = string_to_set(node.right.value)
            states2 = string_to_set(node.left.value) | states1
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down(n, p))
            node.value = str(p)
        
        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cCGS.et_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)
            
def solve_tree_adjacency_list(node):
    if node.left is not None:
        solve_tree(node.left)
    if node.right is not None:
        solve_tree(node.right)
    if node.right is None:   # UNARY OPERATORS: not, globally, next, eventually
        if verify('NOT', node.value):  # e.g. ¬φ
            states = string_to_set(node.left.value)
            all_states = set(cCGS.get_states())
            ris = all_states - states
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('GLOBALLY', node.value):  # e.g. <Jn>Gφ
            #<Jn>falseRφ
            n = int(node.value[2:-2])
            states1 = set()
            states2 = string_to_set(node.left.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down_variant(n, p))
                i+=1
            node.value = str(p)

        elif verify('DEMONIC', node.value) and verify('NEXT', node.value):  # e.g. <Jn>Xφ
            n = int(node.value[2:-2])
            states = string_to_set(node.left.value)
            ris = triangle_down_variant(n, states)
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('EVENTUALLY', node.value):  # e.g. <Jn>Fφ
            #<Jn>trueUϕ.
            n = int(node.value[2:-2])
            states1 = set(cCGS.get_states())
            states2 = string_to_set(node.left.value)
            p = set()
            t = states2
            while t != p:
                p = t
                t = states2 | (states1 & triangle_down_variant(n, p))
            node.value = str(p)

    if node.left is not None and node.right is not None:  # BINARY OPERATORS: or, and, until, implies
        if verify('OR', node.value): # e.g. φ || θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.union(states2)
            node.value = str(ris)

        elif verify('DEMONIC', node.value) and verify('UNTIL', node.value):  # e.g. <Jn>φUθ
            n = int(node.value[2:-2])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set()
            t = states2
            while t != p: 
                p = t
                t = states2 | (states1 & triangle_down_variant(n, p))
            node.value = str(p)
        elif verify('DEMONIC', node.value) and verify('RELEASE', node.value): #e.g. <Jn>φRθ
            n = int(node.value[2:-2])
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down_variant(n, p))
            node.value = str(p)
        elif verify('DEMONIC', node.value) and verify('WEAK', node.value): #e.g. <Jn>φWθ
            #<Jn>(θ R (φ ∨ θ))
            n = int(node.value[2:-2])
            states1 = string_to_set(node.right.value)
            states2 = string_to_set(node.left.value) | states1
            p = set(cCGS.get_states())
            t = states2
            while t != p: 
                p = t
                t = states2 & (states1 | triangle_down_variant(n, p))
            node.value = str(p)
        
        elif verify('AND', node.value):  # e.g. φ && θ
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            ris = states1.intersection(states2)
            node.value = str(ris)

        elif verify('IMPLIES', node.value):  # e.g. φ -> θ
            # p -> q ≡ ¬p ∨ q
            states1 = string_to_set(node.left.value)
            states2 = string_to_set(node.right.value)
            not_states1 = set(cCGS.get_states()).difference(states1)
            ris = not_states1.union(states2)
            node.value = str(ris)



# does the parsing of the model, the formula, builds a tree and then it returns the result of model checking
# function called by front_end_CS
def model_checking(formula, filename):
    global cCGS

    # model parsing
    cCGS = costCGS()
    cCGS.read_file(filename)

    build_pre_set_array()
    if not formula.strip():
        result = {'res': 'Error: formula not entered'}
        return result


    # formula parsing
    res_parsing = do_parsing(formula)
    if res_parsing is None:
        result = {'res': "Syntax Error"}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist"}
        return result

    # model checking
    solve_tree(root)

    # solution
    result = {'res': 'Result: ' + str(root.value)}
    return result
    

def model_checking_test(formula, filename): 
    global last_filename, cCGS
    if not formula.strip():
        result = {'res': 'Error: formula not entered'}
        return result

    cCGS = costCGS()
    if (filename != last_filename):
        # model parsing
        cCGS.read_file(filename)
        build_pre_set_array()
        last_filename = filename
        
    # formula parsing
    res_parsing = do_parsing(formula)
    if res_parsing is None:
        result = {'res': "Syntax Error"}
        return result
    root = build_tree(res_parsing)
    if root is None:
        result = {'res': "Syntax Error: the atom does not exist"}
        return result
        
    # model checking
    time1 = timer()
    solve_tree_adjacency_list(root)
    model_checking_time_adjacency_list = timer() - time1
    result_adjacency = root.value
    root = build_tree(res_parsing)
    # model checking
    time1 = timer()
    solve_tree(root)
    model_checking_time = timer() - time1
    if (root.value != result_adjacency):
        not_equal.append(formula)
    if (root.value != "set()"):
        res_str = root.value
    else:
        res_str = "None"
    result = {'res': res_str, 'time': model_checking_time, 'time_adjacency_list': model_checking_time_adjacency_list}
    return result


import numpy as np
from vitamin_model_checker.models.CGS.CGS import CGS

class capCGS(CGS):
    def init(self):
        super(capCGS, self).init()
        self.capacities_assignment = []
        self.action_capacities = []
        self.capacities = []


    def read_file(self, filename):
        with open(filename, 'r') as f:
            lines = f.readlines()

        self.graph = []
        self.states = []
        self.atomic_propositions = []
        self.matrix_prop = []
        self.initial_state = ''
        self.number_of_agents = ''
        self.capacities_assignment = []
        self.action_capacities = []
        self.capacities = []

        current_section = None
        transition_content = ''
        unknown_transition_content = ''
        name_state_content = ''
        atomic_propositions_content = ''
        labelling_content = ''
        rows_graph = []
        rows_prop = []
        capacities_assignment_content = ''
        action_assign_content = ''
        capacities_content = ''
        

        for line in lines:
            line = line.strip()
            #Section "header"
            if line == 'Transition':
                current_section = 'Transition'
            elif line == 'Unkown_Transition_by':
                current_section = 'Unknown_Transition_by'
            elif line == 'Name_State':
                current_section = 'Name_State'
            elif line == 'Initial_State':
                current_section = 'Initial_State'
            elif line == 'Atomic_propositions':
                current_section = 'Atomic_propositions'
            elif line == 'Labelling':
                current_section = 'Labelling'
            elif line == 'Number_of_agents':
                current_section = 'Number_of_agents'
            elif line == 'Capacities' :
                current_section = 'Capacities'
            elif line == 'Capacities_assignment':
                current_section = 'Capacities_assignment'
            elif line == 'Actions_for_capacities':
                current_section = 'Actions_for_capacities'

            #If not header, then read contents based on what section we are in
            
            elif current_section == 'Transition':
                transition_content += line + '\n'
                values = line.strip().split()
                rows_graph.append(values)
            elif current_section == 'Unknown_Transition_by':
                unknown_transition_content += line + '\n'
            elif current_section == 'Name_State':
                name_state_content += line + '\n'
                values = line.strip().split()
                self.states = np.array(values)
            elif current_section == 'Initial_State':
                self.initial_state = line
            elif current_section == 'Atomic_propositions':
                atomic_propositions_content += line + '\n'
                values = line.strip().split()
                self.atomic_propositions = np.array(values)
            elif current_section == 'Labelling':
                labelling_content += line + '\n'
                values = line.strip().split()
                rows_prop.append(values)
            elif current_section == 'Number_of_agents':
                self.number_of_agents = line
            elif current_section == 'Capacities' :
                capacities_content += line + '\n'
                values = line.strip().split()
                self.capacities = np.array(values)
            elif current_section == 'Capacities_assignment':
                capacities_assignment_content += line + '\n'
                values = line.strip().split()
                self.capacities_assignment.append(values)
            elif current_section == 'Actions_for_capacities':
                action_assign_content += line + '\n'
                values = line.strip().split()
                self.action_capacities.append(values)
            
        actions =[]
        a = 0
        grafo_prov = np.array(rows_graph)
        for row in grafo_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                else:
                    new_row.append(str(item))
                    a = item.split(",")
                    for elem in a :
                        actions.append(elem)
            self.graph.append(new_row)   

        matrix_prop_prov = np.array(rows_prop)
        for row in matrix_prop_prov:
            new_row = []
            for item in row:
                if item == '0':
                    new_row.append(0)
                elif item == '1':
                    new_row.append(1)
                else:
                    new_row.append(str(item))
            self.matrix_prop.append(new_row)
        
    def read_from_model_object(self, model):
        super(capCGS, self).read_from_model_object(model)
        self.capacities_assignment = model.capacities_assignment
        self.action_capacities = model.actions_for_capacities
        self.capacities = np.array(model.capacities)

    def get_capacities_assignment2(self):
        return self.capacities_assignment

    def get_capacities_assignment(self) :
        cap_ass = self.get_capacities_assignment2()
        result = []
        for i in range (1, self.get_number_of_agents()+1) :
            interm = [str(i)]
            cap_ag = cap_ass[i-1]
            for count, value in enumerate(cap_ag) :
                if value == '1' :
                    interm.append(self.get_capacities()[count])
            result.append(interm)

        return result


    def get_action_capacities(self):
        return self.action_capacities

    def get_capacities(self): 
        result = []
        for elem in self.capacities :
            result.append(elem)
        return result


import random
import argparse

def main():
    # Use argparse to parse command-line arguments
    parser = argparse.ArgumentParser(description="Generate a state-transition system with agents and obstacles.")
    parser.add_argument('--agents', type=int, default=2, help='Number of agents')
    parser.add_argument('--rows', type=int, default=2, help='Number of rows in the grid')
    parser.add_argument('--columns', type=int, default=2, help='Number of columns in the grid')
    parser.add_argument('--obstacles', type=int, default=2, help='Number of obstacles in the grid')

    # Parse the arguments
    args = parser.parse_args()

    # Use parsed arguments instead of constants
    NUMBER_OF_AGENTS = args.agents
    NUMBER_OF_ROWS = args.rows
    NUMBER_OF_COLUMNS = args.columns
    NUMBER_OF_OBSTACLES = args.obstacles

    generate_random_CGS(NUMBER_OF_AGENTS, NUMBER_OF_ROWS, NUMBER_OF_COLUMNS, NUMBER_OF_OBSTACLES, 'tmp')

def generate_random_CGS(NUMBER_OF_AGENTS, NUMBER_OF_ROWS, NUMBER_OF_COLUMNS, NUMBER_OF_OBSTACLES, FILENAME):
    propositions = set()

    obstacles = set()
    while len(obstacles) != NUMBER_OF_OBSTACLES:
        obstacles.add(f'o({random.randint(2, NUMBER_OF_ROWS)},{random.randint(1, NUMBER_OF_COLUMNS)})')

    propositions.update(obstacles)

    positions = [(r, c) for r in range(1, NUMBER_OF_ROWS + 1) for c in range(1, NUMBER_OF_COLUMNS + 1)]
    agents_pos = [0] * NUMBER_OF_AGENTS
    states = []
    for _ in range((NUMBER_OF_ROWS * NUMBER_OF_COLUMNS) ** NUMBER_OF_AGENTS):
        state = set()
        for ag in range(1, NUMBER_OF_AGENTS + 1):
            state.add(f'a{ag}{positions[agents_pos[ag - 1]]}')
        for ag in state:
            pos_ag = ag[ag.index('(') + 1:-1].replace(' ', '')
            ok = True
            for obs in obstacles:
                pos_obs = obs[obs.index('(') + 1:-1].replace(' ', '')
                if pos_ag == pos_obs:
                    ok = False
                    break
            if not ok:
                break
        if ok:
            state.update(obstacles)
            states.append(state)
            propositions.update(state)
        for i in range(NUMBER_OF_AGENTS):
            agents_pos[i] = (agents_pos[i] + 1) % len(positions)
            if agents_pos[i] != 0:
                break

    initial_state = f's{random.randint(1, len(states) - 1)}'
    states_names = ' '.join([f's{i}' for i in range(0, len(states))])
    propositions = list(propositions)
    atomic_propositions = ' '.join([prop.replace('(', '').replace(')', '').replace(',', '_').replace(' ', '') for prop in propositions])
    labelling = ''
    for state in states:
        for prop in propositions:
            if prop in state:
                labelling += '1 '
            else:
                labelling += '0 '
        labelling = labelling[:-1]
        labelling += '\n'
    labelling = labelling[:-1]
    transitions = ''
    for from_state in states:
        for to_state in states:
            if from_state == to_state:
                transitions += 'i' * NUMBER_OF_AGENTS + ' '
            else:
                transition = ''
                ok = True
                for ag in range(1, NUMBER_OF_AGENTS + 1):
                    from_pos = [prop for prop in from_state if prop.startswith(f'a{ag}')][0]
                    from_pos = [int(pos.strip()) for pos in from_pos[from_pos.index('(') + 1:-1].split(',')]
                    to_pos = [prop for prop in to_state if prop.startswith(f'a{ag}')][0]
                    to_pos = [int(pos.strip()) for pos in to_pos[to_pos.index('(') + 1:-1].split(',')]
                    diff = (from_pos[0] - to_pos[0], from_pos[1] - to_pos[1])
                    action = None
                    if diff[0] == 0 and diff[1] == 0:
                        action = 'i'
                    if diff[0] == 0 and diff[1] == 1:
                        action = 'W'
                    if diff[0] == 1 and diff[1] == 0:
                        action = 'S'
                    if diff[0] == 0 and diff[1] == -1:
                        action = 'E'
                    if diff[0] == -1 and diff[1] == 0:
                        action = 'N'
                    if not action:
                        transition = '0'
                        break
                    else:
                        transition += action
                transitions += transition + ' '
        transitions += '\n'
    transitions = transitions[:-1]
    res = f'''Transition
{transitions}
Name_State
{states_names}
Initial_State
{initial_state}
Atomic_propositions
{atomic_propositions}
Labelling
{labelling}
Number_of_agents
{NUMBER_OF_AGENTS}
'''
    with open(FILENAME, 'w') as file:
        file.write(res)

if __name__ == "__main__":
    main()



from vitamin_model_checker.model_checker_interface.explicit.CTL import model_checking
from vitamin_model_checker.models.CGS import *

#path destination for new updated input file
pruned_model_file = './tmp.txt'

def modify_matrix(graph, label_matrix, states, action, agent_index, agents):
    #print(f"states:{states}")
    new_graph = [row.copy() for row in graph]
    rows_modified = [False] * len(new_graph)
    for i, row in enumerate(new_graph):
        row_modified = False
        for j, elem in enumerate(row):
            if label_matrix[i][j] in states:
                if isinstance(elem, str) and elem != '*':
                    elem_parts = elem.split(',')
                    new_elem_parts = []
                    for part in elem_parts:
                        part_list = list(part)
                        agent_matches = False
                        if part_list[agents[agent_index-1]-1] == 'I' or part_list[agents[agent_index-1]-1] == action:
                            agent_matches = True
                        if agent_matches:
                            new_elem_parts.append(part)
                    new_elem = ','.join(new_elem_parts)
                    new_graph[i][j] = new_elem if new_elem else 0
                    row_modified = True
        rows_modified[i] = row_modified
    return new_graph

def process_transition_matrix_data(cgs, model, agents,  *strategies):
    graph = cgs.get_graph()
    label_matrix = cgs.create_label_matrix(graph)
    print(f"initial transition matrix: {graph}")
    #print(f"associated labelling matrix:{label_matrix}")
    actions_per_agent = cgs.get_actions(agents)
    agent_actions = {}
    for i, agent_key in enumerate(actions_per_agent.keys()):
        agent_actions[f"actions_{agent_key}"] = actions_per_agent[agent_key]

    for agent_key in agent_actions:
        print(agent_key)
        print(agent_actions[agent_key])

    for strategy_index, strategy in enumerate(strategies, start=1):
        state_sets = set()
        temp = set()
        for iteration, (condition, action) in enumerate(strategy['condition_action_pairs']):
            #print(f"condition {condition}")
            states = model_checking(condition, model)
            #print(f" result: m_checking{states}, statesres {states['res']}")
            state_set = eval(states['res'].split(': ')[1])

            if iteration > 0:
                if (state_set):
                    temp = state_sets
                    state_sets = state_set - temp
                else:
                    state_sets = set(cgs.get_states())
                    action = "I"
                #print(f"state_set: {state_sets} con iteration {iteration} action {action}")
                graph = modify_matrix(graph, label_matrix, state_sets, action, strategy_index, agents)
                #print(f"second iteration modify_matrix for agent {strategy_index}")
                print(f'new transition matrix: {graph} modified by agent {strategy_index}')
            else:
                if (state_set):
                    state_sets = state_set
                else:
                    state_sets = set(cgs.get_states())
                    action = "I"
                #print(f"state_set: {state_sets}, iteration {iteration} action {action}")
                graph = modify_matrix(graph, label_matrix, state_sets, action, strategy_index, agents)
                #print(f"First iteration of modify_matrix for agent {strategy_index}")
                print(f'new transition matrix: {graph} modified by agent {strategy_index}')

    return graph

def pruning(cgs, model, agents, formula, current_agents):
    cgs1 = CGS()
    cgs1.read_file(model)
    cgs1.graph = process_transition_matrix_data(cgs, model, agents,  *current_agents)
    cgs1.matrixParser(cgs.get_number_of_agents())
    cgs1.write_updated_file(model, cgs1.graph, pruned_model_file)
    result = model_checking(formula, pruned_model_file)

    if (result['initial_state'] == 'Initial state s0: True'):
        #print(result)
        return True